from enum import StrEnum


class StandardIcon(StrEnum):
    ADVANCED = "advanced"
    ANGLE_DOWN = "angleDown"
    ANGLE_LEFT = "angleLeft"
    ANGLE_RIGHT = "angleRight"
    ANGLES_DOWN = "anglesDown"
    ANGLES_LEFT = "anglesLeft"
    ANGLES_RIGHT = "anglesRight"
    ANGLES_UP = "anglesUp"
    ANGLE_UP = "angleUp"
    ARROW_DOWN = "arrowDown"
    ARROW_LEFT = "arrowLeft"
    ARROW_RIGHT = "arrowRight"
    ARROW_UP = "arrowUp"
    AWARD = "award"
    BACKGROUND = "background"
    BELL = "bell"
    BELL_SLASH = "bellSlash"
    BOOKMARK = "bookmark"
    BOOKMARK_HOLLOW = "bookmarkHollow"
    CHECK = "check"
    CLOCK = "clock"
    COMMENT = "comment"
    CROWN = "crown"
    DELETE = "delete"
    EDIT = "edit"
    EFFECT = "effect"
    ENGINE = "engine"
    ENVELOPE = "envelope"
    ENVELOPE_OPEN = "envelopeOpen"
    GLOBE = "globe"
    HEART = "heart"
    HEART_HOLLOW = "heartHollow"
    HIDE = "hide"
    INFORMATION = "information"
    LEVEL = "level"
    LOCK = "lock"
    MEDAL = "medal"
    MESSAGE = "message"
    MINUS = "minus"
    OPTIONS = "options"
    PARTICLE = "particle"
    PIN = "pin"
    PLAYER = "player"
    PLAYLIST = "playlist"
    PLUS = "plus"
    POST = "post"
    QUESTION = "question"
    RANKING = "ranking"
    REPLAY = "replay"
    REPLY = "reply"
    RESTORE = "restore"
    ROOM = "room"
    SEARCH = "search"
    SETTINGS = "settings"
    SHOW = "show"
    SHUFFLE = "shuffle"
    SKIN = "skin"
    STAR = "star"
    STAR_HALF = "starHalf"
    STAR_HOLLOW = "starHollow"
    STOPWATCH = "stopwatch"
    TAG = "tag"
    THUMBS_DOWN = "thumbsDown"
    THUMBS_DOWN_HOLLOW = "thumbsDownHollow"
    THUMBS_UP = "thumbsUp"
    THUMBS_UP_HOLLOW = "thumbsUpHollow"
    TROPHY = "trophy"
    UNLOCK = "unlock"
    X_MARK = "xMark"



from __future__ import annotations

from collections.abc import Callable
from enum import Enum
from types import FunctionType, MethodType, NoneType, NotImplementedType
from typing import TYPE_CHECKING, Annotated, Any, Literal, TypeVar, get_origin, overload

if TYPE_CHECKING:
    from sonolus.script.internal.value import Value


@overload
def meta_fn[T: Callable](fn: T) -> T: ...


@overload
def meta_fn[T: Callable]() -> Callable[[T], T]: ...


def meta_fn(fn=None):
    # noinspection PyShadowingNames
    def decorator(fn):
        fn._meta_fn_ = True
        return fn

    if fn is None:
        return decorator
    return decorator(fn)


def validate_value(value: Any) -> Value:
    result = try_validate_value(value)
    if result is None:
        raise TypeError(f"Unsupported value: {value!r}")
    return result


def try_validate_value(value: Any) -> Value | None:
    from sonolus.script.comptime import Comptime
    from sonolus.script.globals import GlobalPlaceholder
    from sonolus.script.internal.generic import PartialGeneric
    from sonolus.script.internal.value import Value
    from sonolus.script.num import Num

    match value:
        case Enum():
            return validate_value(value.value)
        case Value():
            return value
        case type():
            if value in {int, float, bool}:
                return Comptime.accept_unchecked(Num)
            return Comptime.accept_unchecked(value)
        case int() | float() | bool():
            return Num._accept_(value)
        case tuple():
            return Comptime.accept_unchecked(tuple(validate_value(v) for v in value))
        case dict():
            return Comptime.accept_unchecked({validate_value(k)._as_py_(): validate_value(v) for k, v in value.items()})
        case PartialGeneric() | TypeVar() | FunctionType() | MethodType() | NotImplementedType() | str() | NoneType():
            return Comptime.accept_unchecked(value)
        case other_type if get_origin(value) in {Literal, Annotated}:
            return Comptime.accept_unchecked(other_type)
        case GlobalPlaceholder():
            return value.get()
        case comptime_value if getattr(comptime_value, "_is_comptime_value_", False):
            return Comptime.accept_unchecked(comptime_value)
        case _:
            return None



from datetime import timedelta
from itertools import starmap

from hypothesis import given, settings
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.containers import ArrayMap, Pair
from tests.script.conftest import validate_dual_run

ints = st.integers(min_value=-999, max_value=999)
maps = st.dictionaries(ints, ints, min_size=1, max_size=20)


@st.composite
def map_and_key(draw):
    values = draw(maps)
    key = draw(st.sampled_from(list(values.keys())))
    return values, key


@st.composite
def map_and_missing_key(draw):
    values = draw(maps)
    key = draw(ints.filter(lambda x: x not in values))
    return values, key


@given(map_and_key())
@settings(deadline=timedelta(milliseconds=500))
def test_insertion(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        return am[key]

    assert validate_dual_run(fn) == values[key]


@given(map_and_key(), ints)
@settings(deadline=timedelta(milliseconds=500))
def test_update(args, new_value):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        am[key] = new_value
        return am[key]

    assert validate_dual_run(fn) == new_value



from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.range import Range
from tests.script.conftest import validate_dual_run


@given(n=st.integers(min_value=0, max_value=100))
def test_basic_range_iteration(n):
    def fn():
        total = 0
        for i in Range(n):
            total += i
        return total

    expected = sum(range(n))
    result = validate_dual_run(fn)
    assert result == expected


@given(start=st.integers(-100, 100), stop=st.integers(-100, 100))
def test_range_iteration_with_start(start, stop):
    def fn():
        total = 0
        for i in Range(start, stop):
            total += i
        return total

    expected = sum(range(start, stop))
    result = validate_dual_run(fn)
    assert result == expected


@given(
    start=st.integers(-100, 100),
    stop=st.integers(-100, 100),
    step=st.integers(-10, 10).filter(lambda x: x != 0),
)
def test_range_iteration_with_step(start, stop, step):
    def fn():
        total = 0
        for i in Range(start, stop, step):
            total += i
        return total

    expected = sum(range(start, stop, step))
    result = validate_dual_run(fn)
    assert result == expected


@given(
    start=st.integers(-100, 100),
    stop=st.integers(-100, 100),
    step=st.integers(-10, -1),
)
def test_range_iteration_with_negative_step(start, stop, step):
    def fn():
        total = 0
        for i in Range(start, stop, step):
            total += i
        return total

    expected = sum(range(start, stop, step))
    result = validate_dual_run(fn)
    assert result == expected


@given(
    start=st.integers(-100, 100),
    stop=st.integers(-100, 100),
    step=st.integers(-10, 10).filter(lambda x: x != 0),
    value=st.integers(-200, 200),
)
def test_range_contains(start, stop, step, value):
    def fn():
        return value in Range(start, stop, step)

    expected = value in range(start, stop, step)
    result = validate_dual_run(fn)
    assert result == expected


@given(
    start=st.integers(-100, 100),
    stop=st.integers(-100, 100),
    step=st.integers(-10, 10).filter(lambda x: x != 0),
)
def test_range_size(start, stop, step):
    def fn():
        return Range(start, stop, step).size()

    expected = len(range(start, stop, step))
    result = validate_dual_run(fn)
    assert result == expected



from datetime import timedelta

from hypothesis import given, settings
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.containers import VarArray
from sonolus.script.debug import assert_true
from tests.script.conftest import validate_dual_run

ints = st.integers(min_value=-999, max_value=999)
lists = st.lists(ints, min_size=1, max_size=20)
sets = st.sets(ints, min_size=1, max_size=20)


@st.composite
def list_and_index(draw):
    values = draw(lists)
    index = draw(st.integers(min_value=0, max_value=len(values) - 1))
    return values, index


@st.composite
def list_and_insert_index(draw):
    values = draw(lists)
    index = draw(st.integers(min_value=0, max_value=len(values)))
    return values, index


@st.composite
def set_and_present_value(draw):
    values = draw(lists)
    index = draw(st.integers(min_value=0, max_value=len(values) - 1))
    return set(values), values[index]


@st.composite
def set_and_missing_value(draw):
    values = draw(sets)
    missing = draw(ints.filter(lambda x: x not in values))
    return values, missing


def test_var_array_insertion():
    def fn():
        values = Array(2, 4, 6, 8)
        va = VarArray[int, 4].new()
        for i, v in enumerate(values):
            va.append(v)
            assert_true(va[i] == v)
            assert_true(len(va) == i + 1)
        return va

    assert list(validate_dual_run(fn)) == [2, 4, 6, 8]


def test_var_array_pop():
    def fn():
        va = VarArray[int, 4].new()
        va.append(2)
        va.append(4)
        va.append(6)
        va.append(8)
        assert_true(len(va) == 4)
        assert_true(va.pop() == 8)
        assert_true(len(va) == 3)
        assert_true(va.pop(1) == 4)
        return va

    assert list(validate_dual_run(fn)) == [2, 6]


def test_var_array_insert():
    def fn():
        va = VarArray[int, 4].new()
        va.append(2)
        va.append(6)
        va.insert(1, 4)
        va.insert(3, 8)
        return va

    assert list(validate_dual_run(fn)) == [2, 4, 6, 8]


def test_var_array_extend():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(2, 4, 6, 8))
        return va

    assert list(validate_dual_run(fn)) == [2, 4, 6, 8]


@given(list_and_index())
@settings(deadline=timedelta(milliseconds=500))
def test_var_array_pop_insert_round_trip(args):
    values_list, index = args

    values = Array(*values_list)
    value_count = len(values_list)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(len(va) == value_count)
        pop_result = va.pop(index)
        assert_true(len(va) == value_count - 1)
        assert_true(pop_result == values[index])
        va.insert(index, pop_result)
        assert_true(len(va) == value_count)
        return va

    assert list(validate_dual_run(fn)) == values_list


@given(list_and_insert_index())
@settings(deadline=timedelta(milliseconds=500))
def test_var_array_insert_pop_round_trip(args):
    values_list, index = args

    values = Array(*values_list)
    value_count = len(values_list)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(len(va) == value_count)
        insert_value = 100
        va.insert(index, insert_value)
        assert_true(len(va) == value_count + 1)
        assert_true(va[index] == insert_value)
        pop_result = va.pop(index)
        assert_true(len(va) == value_count)
        assert_true(pop_result == insert_value)
        return va

    assert list(validate_dual_run(fn)) == values_list


@given(list_and_index())
@settings(deadline=timedelta(milliseconds=500))
def test_remove_present_value(args):
    values_list, index = args

    values = Array(*values_list)
    value_count = len(values_list)
    to_remove = values[index]
    expected = list(values)
    expected.remove(to_remove)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(va.remove(to_remove))
        return va

    assert list(validate_dual_run(fn)) == expected


@given(lists)
@settings(deadline=timedelta(milliseconds=500))
def test_remove_missing_value(values_list):
    values = Array(*values_list)
    value_count = len(values_list)
    to_remove = max(values) + 1
    expected = list(values)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(not va.remove(to_remove))
        return va

    assert list(validate_dual_run(fn)) == expected


def test_clear():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(2, 4, 6, 8))
        va.clear()
        return va

    assert list(validate_dual_run(fn)) == []


@given(set_and_present_value())
@settings(deadline=timedelta(milliseconds=500))
def test_set_add_present(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(not va.set_add(value))
        return va

    # Use sorted to check that there are no duplicates
    assert sorted(validate_dual_run(fn)) == sorted(values)


@given(set_and_missing_value())
@settings(deadline=timedelta(milliseconds=500))
def test_set_add_missing(args):
    value_set, missing = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(va.set_add(missing))
        return va

    # Use sorted to check that there are no duplicates
    assert sorted(validate_dual_run(fn)) == sorted([*list(values), missing])


def test_set_add_full():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(2, 4, 6, 8))
        assert_true(not va.set_add(2))
        assert_true(not va.set_add(10))
        return va

    assert list(validate_dual_run(fn)) == [2, 4, 6, 8]


@given(set_and_present_value())
@settings(deadline=timedelta(milliseconds=500))
def test_set_remove_present(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(va.remove(value))
        return va

    assert sorted(validate_dual_run(fn)) == sorted(value_set - {value})


@given(set_and_missing_value())
@settings(deadline=timedelta(milliseconds=500))
def test_set_remove_missing(args):
    value_set, missing = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(not va.remove(missing))
        return va

    assert sorted(validate_dual_run(fn)) == sorted(value_set)


@given(set_and_missing_value())
@settings(deadline=timedelta(milliseconds=500))
def test_set_add_remove_round_trip(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(va.set_add(value))
        assert_true(va.remove(value))
        return va

    assert sorted(validate_dual_run(fn)) == sorted(value_set)


@given(set_and_present_value())
@settings(deadline=timedelta(milliseconds=500))
def test_set_remove_add_round_trip(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(va.remove(value))
        assert_true(not va.remove(value))
        assert_true(va.set_add(value))
        return va

    assert sorted(validate_dual_run(fn)) == sorted(value_set)



from collections.abc import Iterable
from typing import overload

from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.iterator import ArrayLike, Enumerator, SonolusIterator
from sonolus.script.math import MATH_BUILTIN_IMPLS
from sonolus.script.num import Num
from sonolus.script.range import Range


@meta_fn
def _isinstance(value, type_):
    value = validate_value(value)
    type_ = validate_value(type_)._as_py_()
    return validate_value(isinstance(value, type_))


@meta_fn
def _len(value):
    from sonolus.backend.visitor import compile_and_call

    value = validate_value(value)
    if not hasattr(value, "__len__"):
        raise TypeError(f"object of type '{type(value).__name__}' has no len()")
    return compile_and_call(value.__len__)


@meta_fn
def _enumerate(iterable, start=0):
    from sonolus.backend.visitor import compile_and_call

    iterable = validate_value(iterable)
    if not hasattr(iterable, "__iter__"):
        raise TypeError(f"'{type(iterable).__name__}' object is not iterable")
    if isinstance(iterable, ArrayLike):
        return compile_and_call(iterable.enumerate, start)
    else:
        iterator = compile_and_call(iterable.__iter__)
        if not isinstance(iterator, SonolusIterator):
            raise TypeError("Only subclasses of SonolusIterator are supported as iterators")
        return Enumerator(0, start, iterator)


@meta_fn
def _abs(value):
    from sonolus.backend.visitor import compile_and_call

    value = validate_value(value)
    if not hasattr(value, "__abs__"):
        raise TypeError(f"bad operand type for abs(): '{type(value).__name__}'")
    return compile_and_call(value.__abs__)


@overload
def _max[T](iterable: Iterable[T]) -> T: ...


@overload
def _max[T](a: T, b: T, *args: T) -> T: ...


@meta_fn
def _max(*args):
    from sonolus.backend.visitor import compile_and_call

    args = tuple(validate_value(arg) for arg in args)
    if len(args) == 0:
        raise ValueError("Expected at least one argument to max")
    elif len(args) == 1:
        (iterable,) = args
        if isinstance(iterable, ArrayLike):
            return iterable.max()
        else:
            raise TypeError(f"Unsupported type: {type(iterable)} for max")
    else:
        if not all(isinstance(arg, Num) for arg in args):
            raise TypeError("Arguments to max must be numbers")
        if ctx():
            result = compile_and_call(_max2, args[0], args[1])
            for arg in args[2:]:
                result = compile_and_call(_max2, result, arg)
            return result
        else:
            return max(arg._as_py_() for arg in args)


def _max2(a, b):
    if a > b:
        return a
    else:
        return b


@overload
def _min[T](iterable: Iterable[T]) -> T: ...


@overload
def _min[T](a: T, b: T, *args: T) -> T: ...


@meta_fn
def _min(*args):
    from sonolus.backend.visitor import compile_and_call

    args = tuple(validate_value(arg) for arg in args)
    if len(args) == 0:
        raise ValueError("Expected at least one argument to min")
    elif len(args) == 1:
        (iterable,) = args
        if isinstance(iterable, ArrayLike):
            return iterable.min()
        else:
            raise TypeError(f"Unsupported type: {type(iterable)} for min")
    else:
        if not all(isinstance(arg, Num) for arg in args):
            raise TypeError("Arguments to min must be numbers")
        if ctx():
            result = compile_and_call(_min2, args[0], args[1])
            for arg in args[2:]:
                result = compile_and_call(_min2, result, arg)
            return result
        else:
            return min(arg._as_py_() for arg in args)


def _min2(a, b):
    if a < b:
        return a
    else:
        return b


BUILTIN_IMPLS = {
    id(isinstance): _isinstance,
    id(len): _len,
    id(enumerate): _enumerate,
    id(abs): _abs,
    id(max): _max,
    id(min): _min,
    id(range): Range,
    **MATH_BUILTIN_IMPLS,
}



from hypothesis import assume, given
from hypothesis import strategies as st

from sonolus.script.interval import Interval, remap, remap_clamped
from tests.script.conftest import implies, is_close, validate_dual_run

ints = st.integers(min_value=-999999, max_value=999999)
floats = st.floats(min_value=-999999, max_value=999999, allow_infinity=False, allow_nan=False)
divisor_floats = floats.filter(lambda x: abs(x) > 1e-6)


@given(floats, floats)
def test_interval_length(start, end):
    def fn():
        interval = Interval(start, end)
        return interval.length

    assert validate_dual_run(fn) == end - start


@given(floats, floats)
def test_interval_mid(start, end):
    def fn():
        interval = Interval(start, end)
        return interval.mid

    assert validate_dual_run(fn) == (start + end) / 2


@given(floats, floats)
def test_interval_is_empty(start, end):
    def fn():
        interval = Interval(start, end)
        return interval.is_empty

    assert validate_dual_run(fn) == (start > end)


@given(floats, floats, floats)
def test_interval_contains_float(start, end, value):
    def fn():
        interval = Interval(start, end)
        return value in interval

    assert validate_dual_run(fn) == (start <= value <= end)


@given(floats, floats, floats, floats)
def test_interval_contains_interval(start, end, other_start, other_end):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        return other in interval

    assert validate_dual_run(fn) == (start <= other_start and other_end <= end)


@given(floats, floats, floats)
def test_interval_add(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval + value

    assert validate_dual_run(fn) == Interval(start + value, end + value)


@given(floats, floats, floats)
def test_interval_sub(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval - value

    assert validate_dual_run(fn) == Interval(start - value, end - value)


@given(floats, floats, floats)
def test_interval_mul(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval * value

    assert validate_dual_run(fn) == Interval(start * value, end * value)


@given(floats, floats, divisor_floats)
def test_interval_truediv(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval / value

    assert validate_dual_run(fn) == Interval(start / value, end / value)


@given(ints, ints, ints.filter(lambda x: x != 0))
def test_interval_floordiv(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval // value

    assert validate_dual_run(fn) == Interval(start // value, end // value)


@given(floats, floats, floats, floats)
def test_interval_intersection(start, end, other_start, other_end):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        return interval & other

    assert validate_dual_run(fn) == Interval(max(start, other_start), min(end, other_end))


@given(floats, floats, floats, floats)
def test_interval_intersection_is_no_longer_than_original(start, end, other_start, other_end):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        intersection = interval & other
        return intersection.length <= interval.length and intersection.length <= other.length

    assert validate_dual_run(fn)


@given(floats, floats, floats, floats, floats)
def test_interval_intersection_consistent_with_contains(start, end, other_start, other_end, value):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        return ((value in interval) and (value in other)) == (value in (interval & other))

    assert validate_dual_run(fn)


@given(floats, floats, floats, floats, floats, floats)
def test_interval_transitive_contains(start1, end1, start2, end2, start3, end3):
    def fn():
        interval1 = Interval(start1, end1)
        interval2 = Interval(start2, end2)
        interval3 = Interval(start3, end3)
        return implies(interval1 in interval2 in interval3, interval1 in interval3)

    assert validate_dual_run(fn)


@given(floats, floats, floats, floats)
def test_interval_intersection_is_in_original(start, end, other_start, other_end):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        intersection = interval & other
        return intersection in interval and intersection in other

    assert validate_dual_run(fn)


@given(floats, floats, floats, floats, floats)
def test_remap_inverse(start, end, other_start, other_end, value):
    assume(abs(end - start) > 1e-6 and abs(other_end - other_start) > 1e-6)

    def fn():
        remapped = remap(start, end, other_start, other_end, value)
        return remap(other_start, other_end, start, end, remapped)

    assert is_close(validate_dual_run(fn), value, abs_tol=1e-4)


@given(floats, floats, floats, floats, floats)
def test_remap_clamped_inverse(start, end, other_start, other_end, value):
    assume(abs(end - start) > 1e-6 and abs(other_end - other_start) > 1e-6)

    def fn():
        remapped = remap_clamped(start, end, other_start, other_end, value)
        return remap_clamped(other_start, other_end, start, end, remapped)

    assert is_close(validate_dual_run(fn), sorted([start, value, end])[1], abs_tol=1e-4)


@given(floats, floats, floats)
def test_lerp_unlerp_inverse(start, end, value):
    assume(abs(end - start) > 1e-6)

    def fn():
        interval = Interval(start, end)
        lerped = interval.lerp(value)
        return interval.unlerp(lerped)

    assert is_close(validate_dual_run(fn), value, abs_tol=1e-4)


@given(floats, floats, floats)
def test_unlerp_lerp_inverse(start, end, value):
    assume(abs(end - start) > 1e-6)

    def fn():
        interval = Interval(start, end)
        unlerped = interval.unlerp(value)
        return interval.lerp(unlerped)

    assert is_close(validate_dual_run(fn), value, abs_tol=1e-4)


@given(floats, floats, floats)
def test_lerp_clamped_unlerp_clamped_inverse(start, end, value):
    assume(abs(end - start) > 1e-6)

    def fn():
        interval = Interval(start, end)
        lerped_clamped = interval.lerp_clamped(value)
        return interval.unlerp_clamped(lerped_clamped)

    assert is_close(validate_dual_run(fn), sorted([0, value, 1])[1], abs_tol=1e-4)


@given(floats, floats, floats)
def test_unlerp_clamped_lerp_clamped_inverse(start, end, value):
    assume(abs(end - start) > 1e-6)

    def fn():
        interval = Interval(start, end)
        unlerped_clamped = interval.unlerp_clamped(value)
        return interval.lerp_clamped(unlerped_clamped)

    assert is_close(validate_dual_run(fn), sorted([start, value, end])[1], abs_tol=1e-4)



from __future__ import annotations

from typing import Protocol

from sonolus.script.record import Record
from sonolus.script.vec import Vec2


class Quad(Record):
    bl: Vec2
    tl: Vec2
    tr: Vec2
    br: Vec2


class Rect(Record):
    t: float
    r: float
    b: float
    l: float  # noqa: E741

    @classmethod
    def from_xywh(cls, x: float, y: float, w: float, h: float) -> Rect:
        return cls(
            t=y + h,
            r=x + w,
            b=y,
            l=x,
        )

    @classmethod
    def from_center(cls, x: float, y: float, w: float, h: float) -> Rect:
        return cls(
            t=y + h / 2,
            r=x + w / 2,
            b=y - h / 2,
            l=x - w / 2,
        )

    @property
    def x(self) -> float:
        return self.l

    @x.setter
    def x(self, value: float):
        self.r += value - self.l
        self.l = value

    @property
    def y(self) -> float:
        return self.t

    @y.setter
    def y(self, value: float):
        self.b += value - self.t
        self.t = value

    @property
    def w(self) -> float:
        return self.r - self.l

    @w.setter
    def w(self, value: float):
        self.r = self.l + value

    @property
    def h(self) -> float:
        return self.t - self.b

    @h.setter
    def h(self, value: float):
        self.t = self.b + value

    @property
    def bl(self) -> Vec2:
        return Vec2(self.l, self.b)

    @property
    def tl(self) -> Vec2:
        return Vec2(self.l, self.t)

    @property
    def tr(self) -> Vec2:
        return Vec2(self.r, self.t)

    @property
    def br(self) -> Vec2:
        return Vec2(self.r, self.b)

    @property
    def center(self) -> Vec2:
        return Vec2((self.l + self.r) / 2, (self.t + self.b) / 2)

    def as_quad(self) -> Quad:
        return Quad(
            bl=self.bl,
            tl=self.tl,
            tr=self.tr,
            br=self.br,
        )

    def scale(self, x: float, y: float | None, /):
        if y is None:
            y = x
        self.l *= x
        self.r *= x
        self.t *= y
        self.b *= y

    def translate(self, x: float, y: float, /):
        self.l += x
        self.r += x
        self.t += y
        self.b += y


class QuadLike(Protocol):
    @property
    def bl(self) -> Vec2: ...

    @property
    def tl(self) -> Vec2: ...

    @property
    def tr(self) -> Vec2: ...

    @property
    def br(self) -> Vec2: ...


def flatten_quad(quad: QuadLike) -> tuple[float, float, float, float, float, float, float, float]:
    return (
        quad.bl.x,
        quad.bl.y,
        quad.tl.x,
        quad.tl.y,
        quad.tr.x,
        quad.tr.y,
        quad.br.x,
        quad.br.y,
    )



import textwrap
from collections import deque
from collections.abc import Iterator

from sonolus.backend.ir import IRConst, IRExpr, IRStmt


class FlowEdge:
    src: "BasicBlock"
    dst: "BasicBlock"
    cond: IRExpr | None

    def __init__(self, src: "BasicBlock", dst: "BasicBlock", cond: IRExpr | None = None):
        self.src = src
        self.dst = dst
        self.cond = cond


class BasicBlock:
    statements: list[IRStmt]
    test: IRExpr
    incoming: set[FlowEdge]
    outgoing: set[FlowEdge]

    def __init__(
        self,
        statements: list[IRStmt] | None = None,
        test: IRExpr | None = None,
        incoming: set[FlowEdge] | None = None,
        outgoing: set[FlowEdge] | None = None,
    ):
        self.statements = statements or []
        self.test = test or IRConst(1)
        self.incoming = incoming or set()
        self.outgoing = outgoing or set()

    def connect_to(self, other: "BasicBlock", cond: IRExpr | None = None):
        edge = FlowEdge(self, other, cond)
        self.outgoing.add(edge)
        other.incoming.add(edge)


def traverse_cfg_preorder(block: BasicBlock) -> Iterator[BasicBlock]:
    visited = set()
    queue = deque([block])
    while queue:
        block = queue.popleft()
        if block in visited:
            continue
        visited.add(block)
        yield block
        for edge in sorted(block.outgoing, key=lambda e: (e.cond is None, e.cond)):
            queue.append(edge.dst)


def cfg_to_mermaid(entry: BasicBlock):
    def pre(s: str):
        return "\"<pre style='text-align: left;'>" + s.replace("\n", "<br/>") + '</pre>"'

    def fmt(nodes):
        if nodes:
            return "\n".join(str(n) for n in nodes)
        else:
            return "{}"

    block_indexes = {block: i for i, block in enumerate(traverse_cfg_preorder(entry))}

    lines = ["Entry([Entry]) --> 0"]
    for block in traverse_cfg_preorder(entry):
        index = block_indexes[block]
        lines.append(f"{index}[{pre(fmt([f'#{index}', *block.statements]))}]")

        outgoing = {edge.cond: edge.dst for edge in block.outgoing}
        match outgoing:
            case {**other} if not other:
                lines.append(f"{index} --> Exit")
            case {None: target, **other} if not other:
                lines.append(f"{index} --> {block_indexes[target]}")
            case {0: f_branch, None: t_branch, **other} if not other:
                lines.append(f"{index}_{{{pre(fmt([block.test]))}}}")
                lines.append(f"{index} --> {index}_")
                lines.append(f"{index}_ --> |true| {block_indexes[t_branch]}")
                lines.append(f"{index}_ --> |false| {block_indexes[f_branch]}")
            case dict() as tgt:
                lines.append(f"{index}_{{{{{pre(fmt([block.test]))}}}}}")
                lines.append(f"{index} --> {index}_")
                for cond, target in tgt.items():
                    lines.append(f"{index}_ --> |{pre(fmt([cond or "default"]))}| {block_indexes[target]}")
    lines.append("Exit([Exit])")

    body = textwrap.indent("\n".join(lines), "    ")
    return f"graph\n{body}"



from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Self, TypeVar, final

from sonolus.backend.place import BlockPlace
from sonolus.script.internal.generic import GenericValue


@final
class _Comptime[T, V](GenericValue):
    _instance: Self | None = None

    def __init__(self):
        super().__init__()
        raise TypeError("Comptime cannot be instantiated")

    @classmethod
    def value(cls):
        _, value = cls._type_args_
        if isinstance(value, Identity):
            return value.value
        return value

    @classmethod
    def _get_parameterized(cls, args: tuple[Any, ...]) -> type[Self]:
        result = super()._get_parameterized(args)
        result._instance = object.__new__(result)
        return result

    @classmethod
    def _size_(cls) -> int:
        return 0

    @classmethod
    def _is_value_type_(cls) -> bool:
        return False

    @classmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        return cls._instance

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        from sonolus.script.internal.impl import validate_value

        value = validate_value(value)
        if not value._is_py_():
            return False
        if cls._type_args_ is None:
            return True
        return value._as_py_() == cls.value()

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        from sonolus.script.internal.impl import validate_value

        if not cls._accepts_(value):
            raise TypeError("Value does not match this Comptime instance")
        # This might not actually return a Comptime instance, but it will be a compile-time constant
        return validate_value(value)

    def _is_py_(self) -> bool:
        return True

    def _as_py_(self) -> Any:
        return self.value()

    @classmethod
    def _from_list_(cls, values: Iterable[float | BlockPlace]) -> Self:
        return cls._instance

    def _to_list_(self) -> list[float | BlockPlace]:
        return []

    @classmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        return []

    def _get_(self) -> Self:
        from sonolus.script.internal.impl import validate_value

        # Converts numbers out of comptime, although _accept_ may end up returning a non-comptime instance anyway
        return validate_value(self.value())

    def _set_(self, value: Self):
        if value is not self:
            raise TypeError("Comptime value cannot be changed")

    def _copy_from_(self, value: Self):
        if value is not self:
            raise TypeError("Comptime value cannot be changed")

    def _copy_(self) -> Self:
        return self

    @classmethod
    def _alloc_(cls) -> Self:
        return cls._instance

    @classmethod
    def _validate__type_args_(cls, args: tuple[Any, ...]) -> tuple[Any, ...]:
        if len(args) == 2:
            _, value = args
            # We want the type to be there for documentation,
            # but not enforced since they might not match up, e.g. a Callable is really FunctionType
            if isinstance(value, TypeVar):
                args = Any, value
            else:
                args = type(value), value
        return super()._validate__type_args_(args)

    @classmethod
    def accept_unchecked(cls, value: Any) -> Self:
        if isinstance(value, dict | tuple):
            args = type(value), Identity(value)
        else:
            args = type(value), value
        if args not in cls._parameterized_:
            cls._parameterized_[args] = cls._get_parameterized(args)
        return cls._parameterized_[args]._instance


class Identity[T]:  # This is to allow accepting potentially unhashable values by using identity comparison
    value: T

    def __init__(self, value: T):
        self.value = value

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return id(self)

    def __str__(self):
        return f"{type(self).__name__}({self.value})"

    def __repr__(self):
        return f"{type(self).__name__}({self.value!r})"


if TYPE_CHECKING:
    type Comptime[T, V] = T | V
else:
    _Comptime.__name__ = "Comptime"
    _Comptime.__qualname__ = "Comptime"
    globals()["Comptime"] = _Comptime



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Collection, Iterator

from sonolus.script.num import Num
from sonolus.script.record import Record
from sonolus.script.values import copy


class SonolusIterator[T](Iterator[T]):
    @abstractmethod
    def has_next(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def next(self) -> T:
        raise NotImplementedError

    def __next__(self) -> T:
        if not self.has_next():
            raise StopIteration
        return self.next()


class ArrayLike[T](Collection):
    @abstractmethod
    def size(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, index: Num) -> T:
        pass

    @abstractmethod
    def __setitem__(self, index: Num, value: T):
        pass

    def __len__(self) -> int:
        return self.size()

    def __iter__(self) -> SonolusIterator[T]:
        return ArrayIterator(0, self)

    def __contains__(self, value: T) -> bool:
        i = 0
        while i < self.size():
            if self[i] == value:
                return True
            i += 1
        return False

    def reversed(self) -> ArrayLike[T]:
        return ArrayReverser(self)

    def iter(self) -> SonolusIterator[T]:
        return self.__iter__()  # noqa: PLC2801

    def enumerate(self, start: Num = 0) -> SonolusIterator[T]:
        return ArrayEnumerator(0, start, self)

    def index_of(self, value: T, start: Num = 0) -> Num:
        i = start
        while i < self.size():
            if self[i] == value:
                return i
            i += 1
        return -1

    def last_index_of(self, value: T) -> Num:
        i = self.size() - 1
        while i >= 0:
            if self[i] == value:
                return i
            i -= 1
        return -1

    def index_of_max(self) -> Num:
        if self.size() == 0:
            return -1
        max_index = 0
        i = 1
        while i < self.size():
            if self[i] > self[max_index]:
                max_index = i
            i += 1
        return max_index

    def index_of_min(self) -> Num:
        if self.size() == 0:
            return -1
        min_index = 0
        i = 1
        while i < self.size():
            if self[i] < self[min_index]:
                min_index = i
            i += 1
        return min_index

    def max(self) -> T:
        return self[self.index_of_max()]

    def min(self) -> T:
        return self[self.index_of_min()]

    def swap(self, i: Num, j: Num):
        temp = copy(self[i])
        self[i] = self[j]
        self[j] = temp

    def sort(self, *, reverse: bool = False):
        if self.size() < 15:
            _insertion_sort(self, 0, self.size(), reverse)
        else:
            _heap_sort(self, 0, self.size(), reverse)


def _insertion_sort[T](array: ArrayLike[T], start: Num, end: Num, reverse: bool):
    i = start + 1
    while i < end:
        value = copy(array[i])
        j = i - 1
        while j >= start and (array[j] > value) != reverse:
            array[j + 1] = array[j]
            j -= 1
        array[j + 1] = value
        i += 1


def _heapify[T](array: ArrayLike[T], end: Num, index: Num, reverse: bool):
    while True:
        left = index * 2 + 1
        right = left + 1
        largest = index
        if left < end and (array[left] > array[largest]) != reverse:
            largest = left
        if right < end and (array[right] > array[largest]) != reverse:
            largest = right
        if largest == index:
            break
        array.swap(index, largest)
        index = largest


# Heap sort is simple to implement iteratively without dynamic memory allocation
def _heap_sort[T](array: ArrayLike[T], start: Num, end: Num, reverse: bool):
    i = end // 2 - 1
    while i >= start:
        _heapify(array, end, i, reverse)
        i -= 1
    i = end - 1
    while i > start:
        array.swap(start, i)
        _heapify(array, i, start, reverse)
        i -= 1


class ArrayIterator[V: ArrayLike](Record, SonolusIterator):
    i: int
    array: V

    def has_next(self) -> bool:
        return self.i < self.array.size()

    def next(self) -> V:
        value = self.array[self.i]
        self.i += 1
        return value


class ArrayReverser[V: ArrayLike](Record, ArrayLike):
    array: V

    def size(self) -> int:
        return self.array.size()

    def __getitem__(self, index: Num) -> V:
        return self.array[self.size() - 1 - index]

    def __setitem__(self, index: Num, value: V):
        self.array[self.size() - 1 - index] = value


class Enumerator[V: SonolusIterator](Record, SonolusIterator):
    i: int
    offset: int
    iterator: V

    def has_next(self) -> bool:
        return self.iterator.has_next()

    def next(self):
        value = self.iterator.next()
        index = self.i + self.offset
        self.i += 1
        return index, value


class ArrayEnumerator[V: ArrayLike](Record, SonolusIterator):
    i: int
    offset: int
    array: V

    def has_next(self) -> bool:
        return self.i < self.array.size()

    def next(self):
        value = self.array[self.i]
        index = self.i + self.offset
        self.i += 1
        return index, value



# fmt: off
from typing import Self

from sonolus.script.graphics import Quad
from sonolus.script.math import cos, sin
from sonolus.script.record import Record
from sonolus.script.vec import Vec2


class Transform2d(Record):
    a00: float
    a01: float
    a02: float
    a10: float
    a11: float
    a12: float
    a20: float
    a21: float

    @classmethod
    def new(cls) -> Self:
        return cls(
            1, 0, 0,
            0, 1, 0,
            0, 0,
        )

    def _compose(self,
        b00: float, b01: float, b02: float,
        b10: float, b11: float, b12: float,
        b20: float, b21: float,
    ) -> Self:
        """Multiply the matrix with another matrix on the left."""
        a00 = self.a00 * b00 + self.a10 * b01 + self.a20 * b02
        a01 = self.a01 * b00 + self.a11 * b01 + self.a21 * b02
        a02 = self.a02 * b00 + self.a12 * b01 + b02
        a10 = self.a00 * b10 + self.a10 * b11 + self.a20 * b12
        a11 = self.a01 * b10 + self.a11 * b11 + self.a21 * b12
        a12 = self.a02 * b10 + self.a12 * b11 + b12
        a20 = self.a00 * b20 + self.a10 * b21 + b20
        a21 = self.a01 * b20 + self.a11 * b21 + b21
        a22 = self.a02 * b20 + self.a12 * b21 + 1
        self.a00 = a00 / a22
        self.a01 = a01 / a22
        self.a02 = a02 / a22
        self.a10 = a10 / a22
        self.a11 = a11 / a22
        self.a12 = a12 / a22
        self.a20 = a20 / a22
        self.a21 = a21 / a22
        return self

    def translate(self, x: float = 0, y: float = 0) -> Self:
        return self._compose(
            1, 0, x,
            0, 1, y,
            0, 0,
        )

    def scale(self, x: float = 1, y: float = 1) -> Self:
        return self._compose(
            x, 0, 0,
            0, y, 0,
            0, 0,
        )

    def rotate(self, angle: float) -> Self:
        c = cos(angle)
        s = sin(angle)
        return self._compose(
            c, -s, 0,
            s, c, 0,
            0, 0,
        )

    def shear_x(self, m: float) -> Self:
        """Shear along the x-axis."""
        return self._compose(
            1, m, 0,
            0, 1, 0,
            0, 0,
        )

    def shear_y(self, m: float) -> Self:
        """Shear along the y-axis."""
        return self._compose(
            1, 0, 0,
            m, 1, 0,
            0, 0,
        )

    def perspective_vanish_y(self, y: float) -> Self:
        return self._compose(
            1, 0, 0,
            0, 1, 0,
            0, 1 / y,
        )

    def compose(self, other: Self) -> Self:
        return self._compose(
            other.a00, other.a01, other.a02,
            other.a10, other.a11, other.a12,
            other.a20, other.a21,
        )

    def transform_vec(self, v: Vec2) -> Vec2:
        x = self.a00 * v.x + self.a01 * v.y + self.a02
        y = self.a10 * v.x + self.a11 * v.y + self.a12
        w = self.a20 * v.x + self.a21 * v.y + 1
        return Vec2(x / w, y / w)

    def transform_quad(self, quad: Quad) -> Quad:
        return Quad(
            bl=self.transform_vec(quad.bl),
            br=self.transform_vec(quad.br),
            tl=self.transform_vec(quad.tl),
            tr=self.transform_vec(quad.tr),
        )



from collections.abc import Callable

from sonolus.backend.finalize import cfg_to_engine_node
from sonolus.backend.flow import BasicBlock
from sonolus.backend.ir import IRConst, IRInstr
from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.backend.optimize import optimize_and_allocate
from sonolus.backend.visitor import compile_and_call
from sonolus.build.node import OutputNodeGenerator
from sonolus.script.archetype import BaseArchetype
from sonolus.script.callbacks import CallbackInfo
from sonolus.script.internal.context import (
    CallbackContextState,
    Context,
    GlobalContextState,
    ReadOnlyMemory,
    context_to_cfg,
    ctx,
    using_ctx,
)
from sonolus.script.num import Num


def compile_mode(
    mode: Mode,
    rom: ReadOnlyMemory,
    archetypes: list[type[BaseArchetype]] | None,
    global_callbacks: list[tuple[CallbackInfo, Callable]] | None,
) -> dict:
    global_state = GlobalContextState(
        mode, {a: i for i, a in enumerate(archetypes)} if archetypes is not None else None, rom
    )
    nodes = OutputNodeGenerator()
    results = {}
    if archetypes is not None:
        archetype_entries = []
        for archetype in archetypes:
            archetype_data = {
                "name": archetype.name,
                "hasInput": archetype.is_scored,
            }
            archetype_data["imports"] = [
                {"name": name, "index": index} for name, index in archetype._imported_keys_.items()
            ]
            if mode == Mode.Play:
                archetype_data["exports"] = [
                    {"name": name, "index": index} for name, index in archetype._exported_keys_.items()
                ]
            for cb_name, cb_info in archetype._supported_callbacks_.items():
                cb = getattr(archetype, cb_name)
                if cb in archetype._default_callbacks_:
                    continue
                cb_order = getattr(cb, "_callback_order_", 0)
                if not cb_info.supports_order and cb_order != 0:
                    raise ValueError(f"Callback '{cb_name}' does not support a non-zero order")
                cfg = callback_to_cfg(global_state, cb, cb_info.name, archetype)
                cfg = optimize_and_allocate(cfg)
                node = cfg_to_engine_node(cfg)
                node_index = nodes.add(node)
                archetype_data[cb_info.name] = {
                    "index": node_index,
                    "order": cb_order,
                }
            archetype_entries.append(archetype_data)
        results["archetypes"] = archetype_entries
    if global_callbacks is not None:
        for cb_info, cb in global_callbacks:
            cfg = callback_to_cfg(global_state, cb, cb_info.name)
            cfg = optimize_and_allocate(cfg)
            node = cfg_to_engine_node(cfg)
            node_index = nodes.add(node)
            results[cb_info.name] = node_index
    results["nodes"] = nodes.get()
    return results


def callback_to_cfg(
    global_state: GlobalContextState, callback: Callable, name: str, archetype: type[BaseArchetype] | None = None
) -> BasicBlock:
    callback_state = CallbackContextState(name)
    context = Context(global_state, callback_state)
    with using_ctx(context):
        if archetype is not None:
            result = compile_and_call(callback, archetype._for_compilation())
        else:
            result = compile_and_call(callback)
        if isinstance(result, Num):
            ctx().add_statements(IRInstr(Op.Break, [IRConst(1), result.ir()]))
    return context_to_cfg(context)



from collections.abc import Iterator
from typing import Self

from sonolus.backend.blocks import Block

type Place = BlockPlace | SSAPlace
type BlockValue = Block | int | TempBlock | Place
type IndexValue = int | Place


class TempBlock:
    name: str
    size: int

    def __init__(self, name: str, size: int = 1):
        self.name = name
        self.size = size

    def __repr__(self):
        return f"TempBlock(name={self.name!r}, size={self.size!r})"

    def __str__(self):
        return f"{self.name}"

    def __getitem__(self, item) -> "BlockPlace":
        return BlockPlace(self, item)

    def __iter__(self) -> "Iterator[BlockPlace]":
        for i in range(self.size):
            yield self[i]

    def __eq__(self, other):
        return isinstance(other, TempBlock) and self.name == other.name and self.size == other.size

    def __hash__(self):
        return hash((self.name, self.size))


class BlockPlace:
    block: BlockValue
    index: IndexValue
    offset: int = 0

    def __init__(self, block: BlockValue, index: IndexValue = 0, offset: int = 0):
        self.block = block
        self.index = index
        self.offset = offset

    def __repr__(self):
        return f"BlockPlace(block={self.block!r}, index={self.index!r}, offset={self.offset!r})"

    def __str__(self):
        if isinstance(self.block, TempBlock) and self.block.size == 1 and self.index == 0 and self.offset == 0:
            return f"{self.block}"
        elif isinstance(self.index, int):
            return f"{self.block}[{self.index + self.offset}]"
        elif self.offset == 0:
            return f"{self.block}[{self.index}]"
        else:
            return f"{self.block}[{self.index} + {self.offset}]"

    def __eq__(self, other):
        return isinstance(other, BlockPlace) and self.block == other.block and self.index == other.index

    def __hash__(self):
        return hash((self.block, self.index))

    def add_offset(self, offset: int) -> Self:
        return BlockPlace(self.block, self.index, self.offset + offset)


class SSAPlace:
    name: str
    num: int

    def __init__(self, name: str, num: int):
        self.name = name
        self.num = num

    def __repr__(self):
        return f"SSAPlace(name={self.name!r}, num={self.num!r})"

    def __str__(self):
        return f"{self.name}.{self.num}"

    def __eq__(self, other):
        return isinstance(other, SSAPlace) and self.name == other.name and self.num == other.num

    def __hash__(self):
        return hash((self.name, self.num))



from dataclasses import dataclass, field
from typing import Literal

UiMetric = Literal[
    "arcade",
    "arcadePercentage",
    "accuracy",
    "accuracyPercentage",
    "life",
    "perfect",
    "perfectPercentage",
    "greatGoodMiss",
    "greatGoodMissPercentage",
    "miss",
    "missPercentage",
    "errorHeatmap",
]

UiJudgmentErrorStyle = Literal[
    "none",
    "plus",
    "minus",
    "arrowUp",
    "arrowDown",
    "arrowLeft",
    "arrowRight",
    "triangleUp",
    "triangleDown",
    "triangleLeft",
    "triangleRight",
]

UiJudgmentErrorPlacement = Literal["both", "left", "right"]

EaseType = Literal[
    "linear",
    "none",
    "inSine",
    "inQuad",
    "inCubic",
    "inQuart",
    "inQuint",
    "inExpo",
    "inCirc",
    "inBack",
    "inElastic",
    "outSine",
    "outQuad",
    "outCubic",
    "outQuart",
    "outQuint",
    "outExpo",
    "outCirc",
    "outBack",
    "outElastic",
    "inOutSine",
    "inOutQuad",
    "inOutCubic",
    "inOutQuart",
    "inOutQuint",
    "inOutExpo",
    "inOutCirc",
    "inOutBack",
    "inOutElastic",
    "outInSine",
    "outInQuad",
    "outInCubic",
    "outInQuart",
    "outInQuint",
    "outInExpo",
    "outInCirc",
    "outInBack",
    "outInElastic",
]


@dataclass
class UiAnimationTween:
    start: float
    end: float
    duration: float
    ease: EaseType

    def to_dict(self):
        return {
            "from": self.start,
            "to": self.end,
            "duration": self.duration,
            "ease": self.ease,
        }


@dataclass
class UiAnimation:
    scale: UiAnimationTween = field(default_factory=lambda: UiAnimationTween(1, 1, 0, "none"))
    alpha: UiAnimationTween = field(default_factory=lambda: UiAnimationTween(1, 0, 0.2, "outCubic"))

    def to_dict(self):
        return {
            "scale": self.scale.to_dict(),
            "alpha": self.alpha.to_dict(),
        }


@dataclass
class UiVisibility:
    scale: float = 1.0
    alpha: float = 1.0

    def to_dict(self):
        return {
            "scale": self.scale,
            "alpha": self.alpha,
        }


@dataclass
class UiConfig:
    scope: str | None = None
    primary_metric: UiMetric = "arcade"
    secondary_metric: UiMetric = "life"
    menu_visibility: UiVisibility = field(default_factory=UiVisibility)
    judgment_visibility: UiVisibility = field(default_factory=UiVisibility)
    combo_visibility: UiVisibility = field(default_factory=UiVisibility)
    primary_metric_visibility: UiVisibility = field(default_factory=UiVisibility)
    secondary_metric_visibility: UiVisibility = field(default_factory=UiVisibility)
    progress_visibility: UiVisibility = field(default_factory=UiVisibility)
    tutorial_navigation_visibility: UiVisibility = field(default_factory=UiVisibility)
    tutorial_instruction_visibility: UiVisibility = field(default_factory=UiVisibility)
    judgment_animation: UiAnimation = field(default_factory=UiAnimation)
    combo_animation: UiAnimation = field(
        default_factory=lambda: UiAnimation(
            scale=UiAnimationTween(1.2, 1, 0.2, "inCubic"), alpha=UiAnimationTween(1, 1, 0, "none")
        )
    )
    judgment_error_style: UiJudgmentErrorStyle = "none"
    judgment_error_placement: UiJudgmentErrorPlacement = "both"
    judgment_error_min: float = 0.0

    def to_dict(self):
        data = {
            "primaryMetric": self.primary_metric,
            "secondaryMetric": self.secondary_metric,
            "menuVisibility": self.menu_visibility.to_dict(),
            "judgmentVisibility": self.judgment_visibility.to_dict(),
            "comboVisibility": self.combo_visibility.to_dict(),
            "primaryMetricVisibility": self.primary_metric_visibility.to_dict(),
            "secondaryMetricVisibility": self.secondary_metric_visibility.to_dict(),
            "progressVisibility": self.progress_visibility.to_dict(),
            "tutorialNavigationVisibility": self.tutorial_navigation_visibility.to_dict(),
            "tutorialInstructionVisibility": self.tutorial_instruction_visibility.to_dict(),
            "judgmentAnimation": self.judgment_animation.to_dict(),
            "comboAnimation": self.combo_animation.to_dict(),
            "judgmentErrorStyle": self.judgment_error_style,
            "judgmentErrorPlacement": self.judgment_error_placement,
            "judgmentErrorMin": self.judgment_error_min,
        }
        if self.scope is not None:
            data["scope"] = self.scope
        return data



from __future__ import annotations

import inspect
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from types import FunctionType
from typing import Annotated, Any, ClassVar, Self, get_origin

from sonolus.backend.ir import IRConst, IRInstr
from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.script.bucket import Bucket, Judgment
from sonolus.script.callbacks import PLAY_CALLBACKS, WATCH_ARCHETYPE_CALLBACKS, CallbackInfo
from sonolus.script.internal.context import ctx
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.generic import validate_concrete_type
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.internal.native import native_call
from sonolus.script.internal.value import Value
from sonolus.script.num import Num
from sonolus.script.pointer import deref
from sonolus.script.record import Record
from sonolus.script.values import zeros

ENTITY_MEMORY_SIZE = 64
ENTITY_DATA_SIZE = 32
ENTITY_SHARED_MEMORY_SIZE = 32


class StorageType(Enum):
    IMPORTED = "imported"
    EXPORTED = "exported"
    MEMORY = "memory"
    SHARED = "shared_memory"


@dataclass
class ArchetypeFieldInfo:
    name: str | None
    storage: StorageType


class ArchetypeField(SonolusDescriptor):
    def __init__(self, name: str, data_name: str, storage: StorageType, offset: int, type_: type[Value]):
        self.name = name
        self.data_name = data_name  # name used in level data
        self.storage = storage
        self.offset = offset
        self.type = type_

    def __get__(self, instance: BaseArchetype, owner):
        if instance is None:
            return self
        result = None
        match self.storage:
            case StorageType.IMPORTED:
                match instance._data_:
                    case ArchetypeSelfData():
                        result = deref(ctx().blocks.EntityData, self.offset, self.type)
                    case ArchetypeReferenceData(index=index):
                        result = deref(ctx().blocks.EntityDataArray, self.offset + index * ENTITY_DATA_SIZE, self.type)
                    case ArchetypeLevelData(values=values):
                        result = values[self.name]
            case StorageType.EXPORTED:
                raise RuntimeError("Exported fields are write-only")
            case StorageType.MEMORY:
                match instance._data_:
                    case ArchetypeSelfData():
                        result = deref(ctx().blocks.EntityMemory, self.offset, self.type)
                    case ArchetypeReferenceData():
                        raise RuntimeError("Entity memory of other entities is not accessible")
                    case ArchetypeLevelData():
                        raise RuntimeError("Entity memory is not available in level data")
            case StorageType.SHARED:
                match instance._data_:
                    case ArchetypeSelfData():
                        result = deref(ctx().blocks.EntitySharedMemory, self.offset, self.type)
                    case ArchetypeReferenceData(index=index):
                        result = deref(
                            ctx().blocks.EntitySharedMemoryArray,
                            self.offset + index * ENTITY_SHARED_MEMORY_SIZE,
                            self.type,
                        )
                    case ArchetypeLevelData():
                        raise RuntimeError("Entity shared memory is not available in level data")
        if result is None:
            raise RuntimeError("Invalid storage type")
        if ctx():
            return result._get_()
        else:
            return result._as_py_()

    def __set__(self, instance: BaseArchetype, value):
        if instance is None:
            raise RuntimeError("Cannot set field on class")
        if not self.type._accepts_(value):
            raise TypeError(f"Expected {self.type}, got {type(value)}")
        target = None
        match self.storage:
            case StorageType.IMPORTED:
                match instance._data_:
                    case ArchetypeSelfData():
                        target = deref(ctx().blocks.EntityData, self.offset, self.type)
                    case ArchetypeReferenceData(index=index):
                        target = deref(ctx().blocks.EntityDataArray, self.offset + index * ENTITY_DATA_SIZE, self.type)
                    case ArchetypeLevelData(values=values):
                        target = values[self.name]
            case StorageType.EXPORTED:
                match instance._data_:
                    case ArchetypeSelfData():
                        if not isinstance(value, self.type):
                            raise TypeError(f"Expected {self.type}, got {type(value)}")
                        for k, v in value._to_flat_dict_(self.data_name).items():
                            index = instance._exported_keys_[k]
                            ctx().add_statements(IRInstr(Op.ExportValue, [IRConst(index), Num._accept_(v).ir()]))
                        return
                    case ArchetypeReferenceData():
                        raise RuntimeError("Exported fields of other entities are not accessible")
                    case ArchetypeLevelData():
                        raise RuntimeError("Exported fields are not available in level data")
            case StorageType.MEMORY:
                match instance._data_:
                    case ArchetypeSelfData():
                        target = deref(ctx().blocks.EntityMemory, self.offset, self.type)
                    case ArchetypeReferenceData():
                        raise RuntimeError("Entity memory of other entities is not accessible")
                    case ArchetypeLevelData():
                        raise RuntimeError("Entity memory is not available in level data")
            case StorageType.SHARED:
                match instance._data_:
                    case ArchetypeSelfData():
                        target = deref(ctx().blocks.EntitySharedMemory, self.offset, self.type)
                    case ArchetypeReferenceData(index=index):
                        target = deref(
                            ctx().blocks.EntitySharedMemoryArray,
                            self.offset + index * ENTITY_SHARED_MEMORY_SIZE,
                            self.type,
                        )
                    case ArchetypeLevelData():
                        raise RuntimeError("Entity shared memory is not available in level data")
        if target is None:
            raise RuntimeError("Invalid storage type")
        value = self.type._accept_(value)
        if self.type._is_value_type_():
            target._set_(value)
        else:
            target._copy_from_(value)


def imported(*, name: str | None = None) -> Any:
    return ArchetypeFieldInfo(name, StorageType.IMPORTED)


def exported(*, name: str | None = None) -> Any:
    return ArchetypeFieldInfo(name, StorageType.EXPORTED)


def entity_memory() -> Any:
    return ArchetypeFieldInfo(None, StorageType.MEMORY)


def shared_memory() -> Any:
    return ArchetypeFieldInfo(None, StorageType.SHARED)


_annotation_defaults: dict[Callable, ArchetypeFieldInfo] = {
    imported: imported(),
    exported: exported(),
    entity_memory: entity_memory(),
    shared_memory: shared_memory(),
}


class StandardImport:
    Beat = Annotated[float, imported(name="#BEAT")]
    Bpm = Annotated[float, imported(name="#BPM")]
    Timescale = Annotated[float, imported(name="#TIMESCALE")]
    Judgment = Annotated[int, imported(name="#JUDGMENT")]
    Accuracy = Annotated[float, imported(name="#ACCURACY")]


def callback[T: Callable](order: int) -> Callable[[T], T]:
    def decorator(func: T) -> T:
        func._callback_order_ = order
        return func

    return decorator


class ArchetypeSelfData:
    pass


class ArchetypeReferenceData:
    index: Num

    def __init__(self, index: Num):
        self.index = index


class ArchetypeLevelData:
    values: dict[str, Value]

    def __init__(self, values: dict[str, Value]):
        self.values = values


type ArchetypeData = ArchetypeSelfData | ArchetypeReferenceData | ArchetypeLevelData


class BaseArchetype:
    _is_comptime_value_ = True

    _supported_callbacks_: ClassVar[dict[str, CallbackInfo]]
    _default_callbacks_: ClassVar[set[Callable]]

    _imported_fields_: ClassVar[dict[str, ArchetypeField]]
    _exported_fields_: ClassVar[dict[str, ArchetypeField]]
    _memory_fields_: ClassVar[dict[str, ArchetypeField]]
    _shared_memory_fields_: ClassVar[dict[str, ArchetypeField]]

    _imported_keys_: ClassVar[dict[str, int]]
    _exported_keys_: ClassVar[dict[str, int]]
    _callbacks_: ClassVar[list[Callable]]
    _data_constructor_signature_: ClassVar[inspect.Signature]
    _spawn_signature_: ClassVar[inspect.Signature]

    _data_: ArchetypeData

    name: ClassVar[str | None] = None
    is_scored: ClassVar[bool] = False

    def __init__(self, *args, **kwargs):
        if ctx():
            raise RuntimeError("The Archetype constructor is only for defining level data")
        bound = self._data_constructor_signature_.bind_partial(*args, **kwargs)
        bound.apply_defaults()
        values = {
            field.name: field.type._accept_(bound.arguments.get(field.name) or zeros(field.type))._get_()
            for field in self._imported_fields_.values()
        }
        self._data_ = ArchetypeLevelData(values=values)

    @classmethod
    def _new(cls):
        return object.__new__(cls)

    @classmethod
    def _for_compilation(cls):
        result = cls._new()
        result._data_ = ArchetypeSelfData()
        return result

    @classmethod
    @meta_fn
    def at(cls, index: Num) -> Self:
        result = cls._new()
        result._data_ = ArchetypeReferenceData(index=index)
        return result

    @classmethod
    @meta_fn
    def id(cls):
        if not ctx():
            raise RuntimeError("Archetype id is only available during compilation")
        result = ctx().global_state.archetypes.get(cls)
        if result is None:
            raise RuntimeError("Archetype is not registered")
        return result

    @classmethod
    @meta_fn
    def spawn(cls, **kwargs):
        if not ctx():
            raise RuntimeError("Spawn is only allowed within a callback")
        archetype_id = cls.id()
        bound = cls._spawn_signature_.bind_partial(**kwargs)
        bound.apply_defaults()
        data = []
        for field in cls._memory_fields_.values():
            data.extend(field.type._accept_(bound.arguments[field.name] or zeros(field.type))._to_list_())
        native_call(Op.Spawn, archetype_id, *(Num(x) for x in data))

    def _level_data_entries(self):
        if not isinstance(self._data_, ArchetypeLevelData):
            raise RuntimeError("Entity is not level data")
        entries = []
        for name, value in self._data_.values.items():
            field_info = self._imported_fields_.get(name)
            for k, v in value._to_flat_dict_(field_info.data_name).items():
                entries.append({"name": k, "value": v})
        return entries

    def __init_subclass__(cls, **kwargs):
        if cls.__module__ == BaseArchetype.__module__:
            if cls._supported_callbacks_ is None:
                raise TypeError("Cannot directly subclass Archetype, use the Archetype subclass for your mode")
            cls._default_callbacks_ = {getattr(cls, cb_info.py_name) for cb_info in cls._supported_callbacks_.values()}
            return
        if getattr(cls, "_callbacks_", None) is not None:
            raise TypeError("Cannot subclass Archetypes")
        if cls.name is None:
            cls.name = cls.__name__
        cls._imported_fields_ = {}
        cls._exported_fields_ = {}
        cls._memory_fields_ = {}
        cls._shared_memory_fields_ = {}
        imported_offset = 0
        exported_offset = 0
        memory_offset = 0
        shared_memory_offset = 0
        for name, value in get_field_specifiers(cls).items():
            if value is ClassVar or get_origin(value) is ClassVar:
                continue
            if get_origin(value) is not Annotated:
                raise TypeError(
                    "Archetype fields must be annotated using imported, exported, entity_memory, or shared_memory"
                )
            field_info = None
            for metadata in value.__metadata__:
                if isinstance(metadata, FunctionType):
                    metadata = _annotation_defaults.get(metadata, metadata)
                if isinstance(metadata, ArchetypeFieldInfo):
                    if field_info is not None:
                        raise TypeError(
                            f"Unexpected multiple field annotations for '{name}', "
                            f"expected exactly one of imported, exported, entity_memory, or shared_memory"
                        )
                    field_info = metadata
            if field_info is None:
                raise TypeError(
                    f"Missing field annotation for '{name}', "
                    f"expected exactly one of imported, exported, entity_memory, or shared_memory"
                )
            field_type = validate_concrete_type(value.__args__[0])
            match field_info.storage:
                case StorageType.IMPORTED:
                    cls._imported_fields_[name] = ArchetypeField(
                        name, field_info.name or name, field_info.storage, imported_offset, field_type
                    )
                    imported_offset += field_type._size_()
                    setattr(cls, name, cls._imported_fields_[name])
                case StorageType.EXPORTED:
                    cls._exported_fields_[name] = ArchetypeField(
                        name, field_info.name or name, field_info.storage, exported_offset, field_type
                    )
                    exported_offset += field_type._size_()
                    setattr(cls, name, cls._exported_fields_[name])
                case StorageType.MEMORY:
                    cls._memory_fields_[name] = ArchetypeField(
                        name, field_info.name or name, field_info.storage, memory_offset, field_type
                    )
                    memory_offset += field_type._size_()
                    setattr(cls, name, cls._memory_fields_[name])
                case StorageType.SHARED:
                    cls._shared_memory_fields_[name] = ArchetypeField(
                        name, field_info.name or name, field_info.storage, shared_memory_offset, field_type
                    )
                    shared_memory_offset += field_type._size_()
                    setattr(cls, name, cls._shared_memory_fields_[name])
        cls._imported_keys_ = {
            name: i
            for i, name in enumerate(
                key for field in cls._imported_fields_.values() for key in field.type._flat_keys_(field.data_name)
            )
        }
        cls._exported_keys_ = {
            name: i
            for i, name in enumerate(
                key for field in cls._exported_fields_.values() for key in field.type._flat_keys_(field.data_name)
            )
        }
        cls._data_constructor_signature_ = inspect.Signature(
            [inspect.Parameter(name, inspect.Parameter.POSITIONAL_OR_KEYWORD) for name in cls._imported_fields_]
        )
        cls._spawn_signature_ = inspect.Signature(
            [inspect.Parameter(name, inspect.Parameter.POSITIONAL_OR_KEYWORD) for name in cls._memory_fields_]
        )
        cls._callbacks_ = []
        for name in cls._supported_callbacks_:
            cb = getattr(cls, name)
            if cb in cls._default_callbacks_:
                continue
            cls._callbacks_.append(cb)


class PlayArchetype(BaseArchetype):
    _supported_callbacks_ = PLAY_CALLBACKS

    def preprocess(self):
        pass

    def spawn_order(self) -> float:
        pass

    def should_spawn(self) -> bool:
        pass

    def initialize(self):
        pass

    def update_sequential(self):
        pass

    def update_parallel(self):
        pass

    def touch(self):
        pass

    def terminate(self):
        pass

    @property
    @meta_fn
    def despawn(self):
        if not ctx():
            raise RuntimeError("Calling despawn is only allowed within a callback")
        match self._data_:
            case ArchetypeSelfData():
                return deref(ctx().blocks.EntityDespawn, 0, Num)
            case _:
                raise RuntimeError("Despawn is only accessible from the entity itself")

    @despawn.setter
    @meta_fn
    def despawn(self, value: bool):
        if not ctx():
            raise RuntimeError("Calling despawn is only allowed within a callback")
        match self._data_:
            case ArchetypeSelfData():
                deref(ctx().blocks.EntityDespawn, 0, Num)._set_(value)
            case _:
                raise RuntimeError("Despawn is only accessible from the entity itself")

    @property
    @meta_fn
    def _info(self):
        if not ctx():
            raise RuntimeError("Calling info is only allowed within a callback")
        match self._data_:
            case ArchetypeSelfData():
                return deref(ctx().blocks.EntityInfo, 0, PlayEntityInfo)
            case ArchetypeReferenceData(index=index):
                return deref(ctx().blocks.EntityInfoArray, index * PlayEntityInfo._size_(), PlayEntityInfo)
            case _:
                raise RuntimeError("Info is only accessible from the entity itself")

    @property
    def index(self) -> int:
        return self._info.index

    @property
    def is_waiting(self) -> bool:
        return self._info.state == 0

    @property
    def is_active(self) -> bool:
        return self._info.state == 1

    @property
    def is_despawned(self) -> bool:
        return self._info.state == 2

    @property
    def life(self) -> ArchetypeLife:
        if not ctx():
            raise RuntimeError("Calling life is only allowed within a callback")
        match self._data_:
            case ArchetypeSelfData() | ArchetypeReferenceData():
                return deref(ctx().blocks.ArchetypeLife, self.id() * ArchetypeLife._size_(), ArchetypeLife)
            case _:
                raise RuntimeError("Life is not available in level data")

    @property
    def result(self) -> PlayEntityInput:
        if not ctx():
            raise RuntimeError("Calling result is only allowed within a callback")
        match self._data_:
            case ArchetypeSelfData():
                return deref(ctx().blocks.EntityInput, 0, PlayEntityInput)
            case _:
                raise RuntimeError("Result is only accessible from the entity itself")


class WatchArchetype(BaseArchetype):
    _supported_callbacks_ = WATCH_ARCHETYPE_CALLBACKS

    def preprocess(self):
        pass

    def spawn_time(self) -> float:
        pass

    def despawn_time(self) -> float:
        pass

    def initialize(self):
        pass

    def update_sequential(self):
        pass

    def update_parallel(self):
        pass

    def terminate(self):
        pass

    @property
    @meta_fn
    def _info(self):
        if not ctx():
            raise RuntimeError("Calling info is only allowed within a callback")
        match self._data_:
            case ArchetypeSelfData():
                return deref(ctx().blocks.EntityInfo, 0, WatchEntityInfo)
            case ArchetypeReferenceData(index=index):
                return deref(ctx().blocks.EntityInfoArray, index * WatchEntityInfo._size_(), PlayEntityInfo)
            case _:
                raise RuntimeError("Info is only accessible from the entity itself")

    @property
    def index(self) -> int:
        return self._info.index

    @property
    def is_active(self) -> bool:
        return self._info.state == 1

    @property
    def life(self) -> ArchetypeLife:
        if not ctx():
            raise RuntimeError("Calling life is only allowed within a callback")
        match self._data_:
            case ArchetypeSelfData() | ArchetypeReferenceData():
                return deref(ctx().blocks.ArchetypeLife, self.id() * ArchetypeLife._size_(), ArchetypeLife)
            case _:
                raise RuntimeError("Life is not available in level data")

    @property
    def result(self) -> WatchEntityInput:
        if not ctx():
            raise RuntimeError("Calling result is only allowed within a callback")
        match self._data_:
            case ArchetypeSelfData():
                return deref(ctx().blocks.EntityInput, 0, WatchEntityInput)
            case _:
                raise RuntimeError("Result is only accessible from the entity itself")

    @property
    def target_time(self) -> float:
        return self.result.target_time

    @target_time.setter
    def target_time(self, value: float):
        self.result.target_time = value


@meta_fn
def entity_info_at(index: Num) -> PlayEntityInfo | WatchEntityInfo | PreviewEntityInfo:
    if not ctx():
        raise RuntimeError("Calling entity_info_at is only allowed within a callback")
    match ctx().global_state.mode:
        case Mode.Play:
            return deref(ctx().blocks.EntityInfoArray, index * PlayEntityInfo._size_(), PlayEntityInfo)
        case Mode.Watch:
            return deref(ctx().blocks.EntityInfoArray, index * WatchEntityInfo._size_(), WatchEntityInfo)
        case Mode.Preview:
            return deref(ctx().blocks.EntityInfoArray, index * PreviewEntityInfo._size_(), PreviewEntityInfo)
        case _:
            raise RuntimeError(f"Entity info is not available in mode '{ctx().global_state.mode}'")


@meta_fn
def archetype_life_of(archetype: type[BaseArchetype] | BaseArchetype) -> ArchetypeLife:
    archetype = validate_value(archetype)
    archetype = archetype._as_py_()
    if not ctx():
        raise RuntimeError("Calling archetype_life_of is only allowed within a callback")
    match ctx().global_state.mode:
        case Mode.Play | Mode.Watch:
            return deref(ctx().blocks.ArchetypeLife, archetype.id() * ArchetypeLife._size_(), ArchetypeLife)
        case _:
            raise RuntimeError(f"Archetype life is not available in mode '{ctx().global_state.mode}'")


class PlayEntityInfo(Record):
    index: int
    archetype_id: int
    state: int


class WatchEntityInfo(Record):
    index: int
    archetype_id: int
    state: int


class PreviewEntityInfo(Record):
    index: int
    archetype_id: int


class ArchetypeLife(Record):
    perfect_increment: Num
    great_increment: Num
    good_increment: Num
    miss_increment: Num

    def update(
        self,
        perfect_increment: Num | None = None,
        great_increment: Num | None = None,
        good_increment: Num | None = None,
        miss_increment: Num | None = None,
    ):
        if perfect_increment is not None:
            self.perfect_increment = perfect_increment
        if great_increment is not None:
            self.great_increment = great_increment
        if good_increment is not None:
            self.good_increment = good_increment
        if miss_increment is not None:
            self.miss_increment = miss_increment


class PlayEntityInput(Record):
    judgment: Judgment
    accuracy: float
    bucket: Bucket
    bucket_value: float


class WatchEntityInput(Record):
    target_time: float
    bucket: Bucket
    bucket_value: float


class EntityRef[A: BaseArchetype](Record):
    index: int

    @classmethod
    def archetype(cls) -> type[A]:
        return cls._get_type_arg_(A)

    def get(self) -> A:
        return self.archetype().at(Num(self.index))



from dataclasses import dataclass
from typing import Annotated, Any, NewType, dataclass_transform, get_origin

from sonolus.backend.ops import Op
from sonolus.script.graphics import QuadLike, flatten_quad
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.internal.native import native_function
from sonolus.script.record import Record
from sonolus.script.vec import Vec2


class Sprite(Record):
    id: int

    @property
    def is_available(self) -> bool:
        return _has_skin_sprite(self.id)

    def draw(self, quad: QuadLike, z: float = 0.0, a: float = 1.0):
        _draw(self.id, *flatten_quad(quad), z, a)

    def draw_curved_b(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        _draw_curved_b(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)

    def draw_curved_t(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        _draw_curved_t(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)

    def draw_curved_l(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        _draw_curved_l(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)

    def draw_curved_r(self, quad: QuadLike, cp: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        _draw_curved_r(self.id, *flatten_quad(quad), z, a, n, *cp.tuple)

    def draw_curved_bt(self, quad: QuadLike, cp1: Vec2, cp2: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        _draw_curved_bt(self.id, *flatten_quad(quad), z, a, n, *cp1.tuple, *cp2.tuple)

    def draw_curved_lr(self, quad: QuadLike, cp1: Vec2, cp2: Vec2, n: float, z: float = 0.0, a: float = 1.0):
        _draw_curved_lr(self.id, *flatten_quad(quad), z, a, n, *cp1.tuple, *cp2.tuple)


@native_function(Op.HasSkinSprite)
def _has_skin_sprite(sprite_id: int) -> bool:
    raise NotImplementedError


@native_function(Op.Draw)
def _draw(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
):
    raise NotImplementedError


@native_function(Op.DrawCurvedB)
def _draw_curved_b(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p: float,
    q: float,
):
    raise NotImplementedError


@native_function(Op.DrawCurvedT)
def _draw_curved_t(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p: float,
    q: float,
):
    raise NotImplementedError


@native_function(Op.DrawCurvedL)
def _draw_curved_l(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p: float,
    q: float,
):
    raise NotImplementedError


@native_function(Op.DrawCurvedR)
def _draw_curved_r(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p: float,
    q: float,
):
    raise NotImplementedError


@native_function(Op.DrawCurvedBT)
def _draw_curved_bt(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p1: float,
    q1: float,
    p2: float,
    q2: float,
):
    raise NotImplementedError


@native_function(Op.DrawCurvedLR)
def _draw_curved_lr(
    sprite_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    z: float,
    a: float,
    n: int,
    p1: float,
    q1: float,
    p2: float,
    q2: float,
):
    raise NotImplementedError


@dataclass
class SkinSprite:
    name: str


def skin_sprite(name: str) -> Any:
    return SkinSprite(name)


type Skin = NewType("Skin", Any)


@dataclass_transform()
def skin[T](cls: type[T]) -> T | Skin:
    if len(cls.__bases__) != 1:
        raise ValueError("Skin class must not inherit from any class (except object)")
    instance = cls()
    names = []
    for i, (name, annotation) in enumerate(get_field_specifiers(cls).items()):
        if get_origin(annotation) is not Annotated:
            raise TypeError(f"Invalid annotation for skin: {annotation}")
        annotation_type = annotation.__args__[0]
        annotation_values = annotation.__metadata__
        if annotation_type is not Sprite:
            raise TypeError(f"Invalid annotation for skin: {annotation}, expected annotation of type Sprite")
        if len(annotation_values) != 1 or not isinstance(annotation_values[0], SkinSprite):
            raise TypeError(f"Invalid annotation for skin: {annotation}, expected a single string annotation value")
        sprite_name = annotation_values[0].name
        names.append(sprite_name)
        setattr(instance, name, Sprite(i))
    instance._sprites_ = names
    instance._is_comptime_value_ = True
    return instance


class StandardSprite:
    NoteHeadNeutral = Annotated[Sprite, skin_sprite("#NOTE_HEAD_NEUTRAL")]
    NoteHeadRed = Annotated[Sprite, skin_sprite("#NOTE_HEAD_RED")]
    NoteHeadGreen = Annotated[Sprite, skin_sprite("#NOTE_HEAD_GREEN")]
    NoteHeadBlue = Annotated[Sprite, skin_sprite("#NOTE_HEAD_BLUE")]
    NoteHeadYellow = Annotated[Sprite, skin_sprite("#NOTE_HEAD_YELLOW")]
    NoteHeadPurple = Annotated[Sprite, skin_sprite("#NOTE_HEAD_PURPLE")]
    NoteHeadCyan = Annotated[Sprite, skin_sprite("#NOTE_HEAD_CYAN")]

    NoteTickNeutral = Annotated[Sprite, skin_sprite("#NOTE_TICK_NEUTRAL")]
    NoteTickRed = Annotated[Sprite, skin_sprite("#NOTE_TICK_RED")]
    NoteTickGreen = Annotated[Sprite, skin_sprite("#NOTE_TICK_GREEN")]
    NoteTickBlue = Annotated[Sprite, skin_sprite("#NOTE_TICK_BLUE")]
    NoteTickYellow = Annotated[Sprite, skin_sprite("#NOTE_TICK_YELLOW")]
    NoteTickPurple = Annotated[Sprite, skin_sprite("#NOTE_TICK_PURPLE")]
    NoteTickCyan = Annotated[Sprite, skin_sprite("#NOTE_TICK_CYAN")]

    NoteTailNeutral = Annotated[Sprite, skin_sprite("#NOTE_TAIL_NEUTRAL")]
    NoteTailRed = Annotated[Sprite, skin_sprite("#NOTE_TAIL_RED")]
    NoteTailGreen = Annotated[Sprite, skin_sprite("#NOTE_TAIL_GREEN")]
    NoteTailBlue = Annotated[Sprite, skin_sprite("#NOTE_TAIL_BLUE")]
    NoteTailYellow = Annotated[Sprite, skin_sprite("#NOTE_TAIL_YELLOW")]
    NoteTailPurple = Annotated[Sprite, skin_sprite("#NOTE_TAIL_PURPLE")]
    NoteTailCyan = Annotated[Sprite, skin_sprite("#NOTE_TAIL_CYAN")]

    NoteConnectionNeutral = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_NEUTRAL")]
    NoteConnectionRed = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_RED")]
    NoteConnectionGreen = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_GREEN")]
    NoteConnectionBlue = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_BLUE")]
    NoteConnectionYellow = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_YELLOW")]
    NoteConnectionPurple = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_PURPLE")]
    NoteConnectionCyan = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_CYAN")]

    NoteConnectionNeutralSeamless = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_NEUTRAL_SEAMLESS")]
    NoteConnectionRedSeamless = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_RED_SEAMLESS")]
    NoteConnectionGreenSeamless = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_GREEN_SEAMLESS")]
    NoteConnectionBlueSeamless = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_BLUE_SEAMLESS")]
    NoteConnectionYellowSeamless = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_YELLOW_SEAMLESS")]
    NoteConnectionPurpleSeamless = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_PURPLE_SEAMLESS")]
    NoteConnectionCyanSeamless = Annotated[Sprite, skin_sprite("#NOTE_CONNECTION_CYAN_SEAMLESS")]

    SimultaneousConnectionNeutral = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_NEUTRAL")]
    SimultaneousConnectionRed = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_RED")]
    SimultaneousConnectionGreen = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_GREEN")]
    SimultaneousConnectionBlue = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_BLUE")]
    SimultaneousConnectionYellow = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_YELLOW")]
    SimultaneousConnectionPurple = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_PURPLE")]
    SimultaneousConnectionCyan = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_CYAN")]

    SimultaneousConnectionNeutralSeamless = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_NEUTRAL_SEAMLESS")]
    SimultaneousConnectionRedSeamless = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_RED_SEAMLESS")]
    SimultaneousConnectionGreenSeamless = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_GREEN_SEAMLESS")]
    SimultaneousConnectionBlueSeamless = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_BLUE_SEAMLESS")]
    SimultaneousConnectionYellowSeamless = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_YELLOW_SEAMLESS")]
    SimultaneousConnectionPurpleSeamless = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_PURPLE_SEAMLESS")]
    SimultaneousConnectionCyanSeamless = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_CONNECTION_CYAN_SEAMLESS")]

    DirectionalMarkerNeutral = Annotated[Sprite, skin_sprite("#DIRECTIONAL_MARKER_NEUTRAL")]
    DirectionalMarkerRed = Annotated[Sprite, skin_sprite("#DIRECTIONAL_MARKER_RED")]
    DirectionalMarkerGreen = Annotated[Sprite, skin_sprite("#DIRECTIONAL_MARKER_GREEN")]
    DirectionalMarkerBlue = Annotated[Sprite, skin_sprite("#DIRECTIONAL_MARKER_BLUE")]
    DirectionalMarkerYellow = Annotated[Sprite, skin_sprite("#DIRECTIONAL_MARKER_YELLOW")]
    DirectionalMarkerPurple = Annotated[Sprite, skin_sprite("#DIRECTIONAL_MARKER_PURPLE")]
    DirectionalMarkerCyan = Annotated[Sprite, skin_sprite("#DIRECTIONAL_MARKER_CYAN")]

    SimultaneousMarkerNeutral = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_MARKER_NEUTRAL")]
    SimultaneousMarkerRed = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_MARKER_RED")]
    SimultaneousMarkerGreen = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_MARKER_GREEN")]
    SimultaneousMarkerBlue = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_MARKER_BLUE")]
    SimultaneousMarkerYellow = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_MARKER_YELLOW")]
    SimultaneousMarkerPurple = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_MARKER_PURPLE")]
    SimultaneousMarkerCyan = Annotated[Sprite, skin_sprite("#SIMULTANEOUS_MARKER_CYAN")]

    StageMiddle = Annotated[Sprite, skin_sprite("#STAGE_MIDDLE")]
    StageLeftBorder = Annotated[Sprite, skin_sprite("#STAGE_LEFT_BORDER")]
    StageRightBorder = Annotated[Sprite, skin_sprite("#STAGE_RIGHT_BORDER")]
    StageTopBorder = Annotated[Sprite, skin_sprite("#STAGE_TOP_BORDER")]
    StageBottomBorder = Annotated[Sprite, skin_sprite("#STAGE_BOTTOM_BORDER")]

    StageLeftBorderSeamless = Annotated[Sprite, skin_sprite("#STAGE_LEFT_BORDER_SEAMLESS")]
    StageRightBorderSeamless = Annotated[Sprite, skin_sprite("#STAGE_RIGHT_BORDER_SEAMLESS")]
    StageTopBorderSeamless = Annotated[Sprite, skin_sprite("#STAGE_TOP_BORDER_SEAMLESS")]
    StageBottomBorderSeamless = Annotated[Sprite, skin_sprite("#STAGE_BOTTOM_BORDER_SEAMLESS")]

    StageTopLeftCorner = Annotated[Sprite, skin_sprite("#STAGE_TOP_LEFT_CORNER")]
    StageTopRightCorner = Annotated[Sprite, skin_sprite("#STAGE_TOP_RIGHT_CORNER")]
    StageBottomLeftCorner = Annotated[Sprite, skin_sprite("#STAGE_BOTTOM_LEFT_CORNER")]
    StageBottomRightCorner = Annotated[Sprite, skin_sprite("#STAGE_BOTTOM_RIGHT_CORNER")]

    Lane = Annotated[Sprite, skin_sprite("#LANE")]
    LaneSeamless = Annotated[Sprite, skin_sprite("#LANE_SEAMLESS")]
    LaneAlternative = Annotated[Sprite, skin_sprite("#LANE_ALTERNATIVE")]
    LaneAlternativeSeamless = Annotated[Sprite, skin_sprite("#LANE_ALTERNATIVE_SEAMLESS")]

    JudgmentLine = Annotated[Sprite, skin_sprite("#JUDGMENT_LINE")]
    NoteSlot = Annotated[Sprite, skin_sprite("#NOTE_SLOT")]
    StageCover = Annotated[Sprite, skin_sprite("#STAGE_COVER")]

    GridNeutral = Annotated[Sprite, skin_sprite("#GRID_NEUTRAL")]
    GridRed = Annotated[Sprite, skin_sprite("#GRID_RED")]
    GridGreen = Annotated[Sprite, skin_sprite("#GRID_GREEN")]
    GridBlue = Annotated[Sprite, skin_sprite("#GRID_BLUE")]
    GridYellow = Annotated[Sprite, skin_sprite("#GRID_YELLOW")]
    GridPurple = Annotated[Sprite, skin_sprite("#GRID_PURPLE")]
    GridCyan = Annotated[Sprite, skin_sprite("#GRID_CYAN")]


@skin
class EmptySkin:
    pass



from enum import IntEnum

from sonolus.backend.mode import Mode
from sonolus.script.array import Array
from sonolus.script.containers import VarArray
from sonolus.script.globals import (
    _level_life,
    _level_score,
    _play_runtime_environment,
    _play_runtime_update,
    _preview_runtime_environment,
    _runtime_background,
    _runtime_particle_transform,
    _runtime_skin_transform,
    _runtime_touch_array,
    _runtime_ui,
    _runtime_ui_configuration,
    _tutorial_runtime_environment,
    _tutorial_runtime_update,
    _watch_runtime_environment,
    _watch_runtime_update,
)
from sonolus.script.graphics import Quad, Rect
from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import meta_fn
from sonolus.script.num import Num
from sonolus.script.record import Record
from sonolus.script.transform import Transform2d
from sonolus.script.vec import Vec2


@_play_runtime_environment
class _PlayRuntimeEnvironment:
    is_debug: bool
    aspect_ratio: float
    audio_offset: float
    input_offset: float
    is_multiplayer: bool


@_watch_runtime_environment
class _WatchRuntimeEnvironment:
    is_debug: bool
    aspect_ratio: float
    audio_offset: float
    input_offset: float
    is_replay: bool


@_preview_runtime_environment
class _PreviewRuntimeEnvironment:
    is_debug: bool
    aspect_ratio: float


@_tutorial_runtime_environment
class _TutorialRuntimeEnvironment:
    is_debug: bool
    aspect_ratio: float
    audio_offset: float


@_play_runtime_update
class _PlayRuntimeUpdate:
    time: float
    delta_time: float
    scaled_time: float
    touch_count: int


@_watch_runtime_update
class _WatchRuntimeUpdate:
    time: float
    delta_time: float
    scaled_time: float
    is_skip: bool


@_tutorial_runtime_update
class _TutorialRuntimeUpdate:
    time: float
    delta_time: float
    navigation_direction: int


class RuntimeUiConfig(Record):
    scale: float
    alpha: float


@_runtime_ui_configuration
class _RuntimeUiConfigs:
    menu: RuntimeUiConfig
    judgment: RuntimeUiConfig
    combo: RuntimeUiConfig
    primary_metric: RuntimeUiConfig
    secondary_metric: RuntimeUiConfig
    progress: RuntimeUiConfig


class HorizontalAlign(IntEnum):
    LEFT = -1
    CENTER = 0
    RIGHT = 1


class RuntimeUiLayout(Record):
    anchor: Vec2
    pivot: Vec2
    dimensions: Vec2
    rotation: float
    alpha: float
    horizontal_align: HorizontalAlign
    background: bool

    def update(
        self,
        anchor: Vec2 | None = None,
        pivot: Vec2 | None = None,
        dimensions: Vec2 | None = None,
        rotation: float | None = None,
        alpha: float | None = None,
        horizontal_align: HorizontalAlign | None = None,
        background: bool | None = None,
    ):
        if anchor is not None:
            self.anchor = anchor
        if pivot is not None:
            self.pivot = pivot
        if dimensions is not None:
            self.dimensions = dimensions
        if rotation is not None:
            self.rotation = rotation
        if alpha is not None:
            self.alpha = alpha
        if horizontal_align is not None:
            self.horizontal_align = horizontal_align
        if background is not None:
            self.background = background


@_runtime_ui
class _RuntimeUi:
    menu: RuntimeUiLayout
    judgment: RuntimeUiLayout
    combo_value: RuntimeUiLayout
    combo_text: RuntimeUiLayout
    primary_metric_bar: RuntimeUiLayout
    primary_metric_value: RuntimeUiLayout
    secondary_metric_bar: RuntimeUiLayout
    secondary_metric_value: RuntimeUiLayout
    progress: RuntimeUiLayout


class Touch(Record):
    id: int
    started: bool
    ended: bool
    time: float
    start_time: float
    position: Vec2
    start_position: Vec2
    delta: Vec2
    velocity: Vec2
    speed: float
    angle: float

    @property
    def total_time(self) -> float:
        return self.time - self.start_time

    @property
    def total_delta(self) -> Vec2:
        return self.position - self.start_position

    @property
    def total_velocity(self) -> Vec2:
        return self.total_delta / self.total_time if self.total_time > 0 else Vec2(0, 0)

    @property
    def total_speed(self) -> float:
        return self.total_velocity.magnitude

    @property
    def total_angle(self) -> float:
        return self.total_delta.angle


@_runtime_touch_array
class _TouchArray:
    touches: Array[Touch, 999]


@_runtime_skin_transform
class _SkinTransform:
    value: Array[Array[float, 4], 4]

    @property
    @meta_fn
    def transform(self) -> Transform2d:
        values = self.value._to_list_()
        return Transform2d._raw(
            a00=Num(values[0 * 4 + 0]),
            a01=Num(values[0 * 4 + 1]),
            a02=Num(values[0 * 4 + 3]),
            a10=Num(values[1 * 4 + 0]),
            a11=Num(values[1 * 4 + 1]),
            a12=Num(values[1 * 4 + 3]),
            a20=Num(values[3 * 4 + 0]),
            a21=Num(values[3 * 4 + 1]),
        )


@_runtime_particle_transform
class _ParticleTransform:
    value: Array[Array[float, 4], 4]

    @property
    @meta_fn
    def transform(self) -> Transform2d:
        values = self.value._to_list_()
        return Transform2d._raw(
            a00=Num(values[0 * 4 + 0]),
            a01=Num(values[0 * 4 + 1]),
            a02=Num(values[0 * 4 + 3]),
            a10=Num(values[1 * 4 + 0]),
            a11=Num(values[1 * 4 + 1]),
            a12=Num(values[1 * 4 + 3]),
            a20=Num(values[3 * 4 + 0]),
            a21=Num(values[3 * 4 + 1]),
        )


@_runtime_background
class _Background:
    value: Quad


@_level_score
class _LevelScore:
    perfect_multiplier: float
    great_multiplier: float
    good_multiplier: float
    consecutive_perfect_multiplier: float
    consecutive_perfect_step: float
    consecutive_perfect_cap: float
    consecutive_great_multiplier: float
    consecutive_great_step: float
    consecutive_great_cap: float
    consecutive_good_multiplier: float
    consecutive_good_step: float
    consecutive_good_cap: float

    def update(
        self,
        perfect_multiplier: float | None = None,
        great_multiplier: float | None = None,
        good_multiplier: float | None = None,
        consecutive_perfect_multiplier: float | None = None,
        consecutive_perfect_step: float | None = None,
        consecutive_perfect_cap: float | None = None,
        consecutive_great_multiplier: float | None = None,
        consecutive_great_step: float | None = None,
        consecutive_great_cap: float | None = None,
        consecutive_good_multiplier: float | None = None,
        consecutive_good_step: float | None = None,
        consecutive_good_cap: float | None = None,
    ):
        if perfect_multiplier is not None:
            self.perfect_multiplier = perfect_multiplier
        if great_multiplier is not None:
            self.great_multiplier = great_multiplier
        if good_multiplier is not None:
            self.good_multiplier = good_multiplier
        if consecutive_perfect_multiplier is not None:
            self.consecutive_perfect_multiplier = consecutive_perfect_multiplier
        if consecutive_perfect_step is not None:
            self.consecutive_perfect_step = consecutive_perfect_step
        if consecutive_perfect_cap is not None:
            self.consecutive_perfect_cap = consecutive_perfect_cap
        if consecutive_great_multiplier is not None:
            self.consecutive_great_multiplier = consecutive_great_multiplier
        if consecutive_great_step is not None:
            self.consecutive_great_step = consecutive_great_step
        if consecutive_great_cap is not None:
            self.consecutive_great_cap = consecutive_great_cap
        if consecutive_good_multiplier is not None:
            self.consecutive_good_multiplier = consecutive_good_multiplier
        if consecutive_good_step is not None:
            self.consecutive_good_step = consecutive_good_step
        if consecutive_good_cap is not None:
            self.consecutive_good_cap = consecutive_good_cap


@_level_life
class _LevelLife:
    consecutive_perfect_increment: float
    consecutive_perfect_step: float
    consecutive_great_increment: float
    consecutive_great_step: float
    consecutive_good_increment: float
    consecutive_good_step: float

    def update(
        self,
        consecutive_perfect_increment: float | None = None,
        consecutive_perfect_step: float | None = None,
        consecutive_great_increment: float | None = None,
        consecutive_great_step: float | None = None,
        consecutive_good_increment: float | None = None,
        consecutive_good_step: float | None = None,
    ):
        if consecutive_perfect_increment is not None:
            self.consecutive_perfect_increment = consecutive_perfect_increment
        if consecutive_perfect_step is not None:
            self.consecutive_perfect_step = consecutive_perfect_step
        if consecutive_great_increment is not None:
            self.consecutive_great_increment = consecutive_great_increment
        if consecutive_great_step is not None:
            self.consecutive_great_step = consecutive_great_step
        if consecutive_good_increment is not None:
            self.consecutive_good_increment = consecutive_good_increment
        if consecutive_good_step is not None:
            self.consecutive_good_step = consecutive_good_step


@meta_fn
def is_debug() -> bool:
    if not ctx():
        return False
    match ctx().global_state.mode:
        case Mode.Play:
            return _PlayRuntimeEnvironment.is_debug
        case Mode.Watch:
            return _WatchRuntimeEnvironment.is_debug
        case Mode.Preview:
            return _PreviewRuntimeEnvironment.is_debug
        case Mode.Tutorial:
            return _TutorialRuntimeEnvironment.is_debug
        case _:
            return False


@meta_fn
def aspect_ratio() -> float:
    if not ctx():
        return 16 / 9
    match ctx().global_state.mode:
        case Mode.Play:
            return _PlayRuntimeEnvironment.aspect_ratio
        case Mode.Watch:
            return _WatchRuntimeEnvironment.aspect_ratio
        case Mode.Preview:
            return _PreviewRuntimeEnvironment.aspect_ratio
        case Mode.Tutorial:
            return _TutorialRuntimeEnvironment.aspect_ratio


@meta_fn
def audio_offset() -> float:
    if not ctx():
        return 0
    match ctx().global_state.mode:
        case Mode.Play:
            return _PlayRuntimeEnvironment.audio_offset
        case Mode.Watch:
            return _WatchRuntimeEnvironment.audio_offset
        case Mode.Tutorial:
            return _TutorialRuntimeEnvironment.audio_offset
        case _:
            return 0


@meta_fn
def input_offset() -> float:
    if not ctx():
        return 0
    match ctx().global_state.mode:
        case Mode.Play:
            return _PlayRuntimeEnvironment.input_offset
        case Mode.Watch:
            return _WatchRuntimeEnvironment.input_offset
        case _:
            return 0


@meta_fn
def is_multiplayer() -> bool:
    if not ctx():
        return False
    match ctx().global_state.mode:
        case Mode.Play:
            return _PlayRuntimeEnvironment.is_multiplayer
        case _:
            return False


@meta_fn
def is_replay() -> bool:
    if not ctx():
        return False
    match ctx().global_state.mode:
        case Mode.Watch:
            return _WatchRuntimeEnvironment.is_replay
        case _:
            return False


@meta_fn
def time() -> float:
    if not ctx():
        return 0
    match ctx().global_state.mode:
        case Mode.Play:
            return _PlayRuntimeUpdate.time
        case Mode.Watch:
            return _WatchRuntimeUpdate.time
        case Mode.Tutorial:
            return _TutorialRuntimeUpdate.time
        case _:
            return 0


@meta_fn
def delta_time() -> float:
    if not ctx():
        return 0
    match ctx().global_state.mode:
        case Mode.Play:
            return _PlayRuntimeUpdate.delta_time
        case Mode.Watch:
            return _WatchRuntimeUpdate.delta_time
        case Mode.Tutorial:
            return _TutorialRuntimeUpdate.delta_time
        case _:
            return 0


@meta_fn
def scaled_time() -> float:
    if not ctx():
        return 0
    match ctx().global_state.mode:
        case Mode.Play:
            return _PlayRuntimeUpdate.scaled_time
        case Mode.Watch:
            return _WatchRuntimeUpdate.scaled_time
        case Mode.Tutorial:
            return _TutorialRuntimeUpdate.time
        case _:
            return 0


@meta_fn
def touches() -> VarArray[Touch, 999]:
    if not ctx():
        return VarArray(0, Array[Touch, 0]())
    match ctx().global_state.mode:
        case Mode.Play:
            return VarArray(_PlayRuntimeUpdate.touch_count, _TouchArray.touches)
        case _:
            return VarArray(0, Array[Touch, 0]())


@meta_fn
def is_skip() -> bool:
    if not ctx():
        return False
    match ctx().global_state.mode:
        case Mode.Watch:
            return _WatchRuntimeUpdate.is_skip
        case _:
            return False


@meta_fn
def navigation_direction() -> int:
    if not ctx():
        return 0
    match ctx().global_state.mode:
        case Mode.Tutorial:
            return _TutorialRuntimeUpdate.navigation_direction
        case _:
            return 0


def skin_transform() -> Transform2d:
    return _SkinTransform.transform


@meta_fn
def set_skin_transform(value: Transform2d):
    _SkinTransform.transform._copy_from_(value)


def particle_transform() -> Transform2d:
    return _ParticleTransform.transform


@meta_fn
def set_particle_transform(value: Transform2d):
    _ParticleTransform.transform._copy_from_(value)


def background() -> Quad:
    return _Background.value


def set_background(value: Quad):
    _Background.value = value


runtime_ui = _RuntimeUi
runtime_ui_configs = _RuntimeUiConfigs


def screen() -> Rect:
    return Rect(t=1, r=aspect_ratio(), b=-1, l=-aspect_ratio())


def level_score() -> _LevelScore:
    return _LevelScore


def level_life() -> _LevelLife:
    return _LevelLife



from collections.abc import Callable

from sonolus.backend.blocks import PlayBlock
from sonolus.backend.finalize import cfg_to_engine_node
from sonolus.backend.interpret import Interpreter
from sonolus.backend.mode import Mode
from sonolus.backend.optimize import optimize_and_allocate
from sonolus.backend.place import BlockPlace
from sonolus.backend.visitor import compile_and_call
from sonolus.build.compile import callback_to_cfg
from sonolus.script.internal.context import GlobalContextState
from sonolus.script.internal.impl import meta_fn
from sonolus.script.num import Num


def compile_fn(callback: Callable):
    global_state = GlobalContextState(Mode.Play)
    return callback_to_cfg(global_state, callback, ""), global_state.rom.values


def validate_dual_run[**P, R](fn: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
    """Runs a function as a regular function and as a compiled function, and checks that the results are the same."""
    exception = None
    regular_result = None
    try:
        regular_result = fn(*args, **kwargs)
    except Exception as e:
        exception = e

    result_type = None

    @meta_fn
    def run_compiled():
        nonlocal result_type
        result = compile_and_call(fn, *args, **kwargs)
        # If terminated, this line won't run
        # We can check whether this value is 1 to see if the compiled function finished without terminating
        Num._from_place_(BlockPlace(-1, 0))._set_(Num(1))
        result_type = type(result)
        target = result_type._from_place_(BlockPlace(-2, 0))
        if result_type._is_value_type_():
            target._set_(result)
        else:
            target._copy_from_(result)
        return result

    cfg, rom_values = compile_fn(run_compiled)
    cfg = optimize_and_allocate(cfg)
    entry = cfg_to_engine_node(cfg)
    interpreter = Interpreter()
    interpreter.blocks[PlayBlock.EngineRom] = rom_values

    num_result = interpreter.run(entry)
    if exception is None:
        if result_type == Num:
            assert num_result == regular_result
        else:
            assert num_result == 0
    compiled_result = result_type._from_list_([interpreter.get(-2, i) for i in range(result_type._size_())])._as_py_()
    compiled_terminated = interpreter.get(-1, 0) != 1

    if exception is not None:
        assert compiled_terminated, "Compiled function should terminate if regular function raises exception"
        raise exception

    assert regular_result == compiled_result

    return regular_result


def compiled_run[**P](fn: Callable[P, Num], *args: P.args, **kwargs: P.kwargs) -> Num:
    """Runs a function as a compiled function and returns the result."""

    @meta_fn
    def run_compiled():
        return compile_and_call(fn, *args, **kwargs)

    cfg, rom_values = compile_fn(run_compiled)
    cfg = optimize_and_allocate(cfg)
    entry = cfg_to_engine_node(cfg)
    interpreter = Interpreter()
    interpreter.blocks[PlayBlock.EngineRom] = rom_values
    return interpreter.run(entry)


def implies(a: bool, b: bool) -> bool:
    if a:
        return b
    return True


def is_close(a: float, b: float, rel_tol: float = 1e-8, abs_tol: float = 1e-8) -> bool:
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)



from os import PathLike
from pathlib import Path

from sonolus.script.engine import Engine
from sonolus.script.level import Level


class Project:
    def __init__(
        self,
        engine: Engine,
        levels: list[Level] | None = None,
        resources: PathLike | None = None,
    ):
        self.engine = engine
        self.levels = levels or []
        self.resources = Path(resources or "resources")



from sonolus.build.engine import package_output
from sonolus.script.level import LevelData


def package_level_data(
    level_data: LevelData,
):
    return package_output(build_level_data(level_data))


def build_level_data(
    level_data: LevelData,
):
    return {
        "bgmOffset": level_data.bgm_offset,
        "entities": [
            {
                "archetype": entity.name,
                "data": entity._level_data_entries(),
            }
            for entity in level_data.entities
        ],
    }



from pathlib import Path

from sonolus.build.collection import Asset, Collection, Srl
from sonolus.build.engine import package_engine
from sonolus.build.level import package_level_data
from sonolus.script.engine import Engine
from sonolus.script.level import Level
from sonolus.script.project import Project

BLANK_PNG = (
    b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x01\x00\x00\x00\x007n\xf9$"
    b"\x00\x00\x00\nIDATx\x01c`\x00\x00\x00\x02\x00\x01su\x01\x18\x00\x00\x00\x00IEND\xaeB`\x82"
)
BLANK_AUDIO = (
    b"RIFF$\x00\x00\x00WAVEfmt \x10\x00\x00\x00\x01\x00\x01\x00D\xac\x00\x00\x88X\x01\x00\x02"
    b"\x00\x10\x00data\x00\x00\x00\x00"
)


def build_project_to_collection(project: Project):
    collection = load_resources_files_to_collection(project.resources)
    add_engine_to_collection(collection, project, project.engine)
    for level in project.levels:
        add_level_to_collection(collection, project, level)
    collection.name = f"{project.engine.name}"
    return collection


def add_engine_to_collection(collection: Collection, project: Project, engine: Engine):
    packaged_engine = package_engine(engine.data)
    item = {
        "name": engine.name,
        "version": engine.version,
        "title": engine.title,
        "subtitle": engine.subtitle,
        "author": engine.author,
        "tags": [],
        "skin": collection.get_item("skins", engine.skin) if engine.skin else collection.get_default_item("skins"),
        "background": collection.get_item("backgrounds", engine.background)
        if engine.background
        else collection.get_default_item("backgrounds"),
        "effect": collection.get_item("effects", engine.effect)
        if engine.effect
        else collection.get_default_item("effects"),
        "particle": collection.get_item("particles", engine.particle)
        if engine.particle
        else collection.get_default_item("particles"),
        "thumbnail": load_resource(collection, engine.thumbnail, project.resources, BLANK_PNG),
        "playData": collection.add_asset(packaged_engine.play_data),
        "watchData": collection.add_asset(packaged_engine.watch_data),
        "previewData": collection.add_asset(packaged_engine.preview_data),
        "tutorialData": collection.add_asset(packaged_engine.tutorial_data),
        "rom": collection.add_asset(packaged_engine.rom),
        "configuration": collection.add_asset(packaged_engine.configuration),
    }
    collection.add_item("engines", engine.name, item)


def add_level_to_collection(collection: Collection, project: Project, level: Level):
    packaged_level_data = package_level_data(level.data)
    item = {
        "name": level.name,
        "version": level.version,
        "rating": level.rating,
        "title": level.title,
        "artists": level.artists,
        "author": level.author,
        "tags": [],
        "engine": collection.get_item("engines", project.engine.name),
        "useSkin": {"useDefault": True},
        "useBackground": {"useDefault": True},
        "useEffect": {"useDefault": True},
        "useParticle": {"useDefault": True},
        "cover": load_resource(collection, level.cover, project.resources, BLANK_PNG),
        "bgm": load_resource(collection, level.bgm, project.resources, BLANK_AUDIO),
        "data": collection.add_asset(packaged_level_data),
    }
    collection.add_item("levels", level.name, item)


def load_resource(collection: Collection, asset: Asset | None, base_path: Path, default: bytes) -> Srl:
    if asset is None:
        return collection.add_asset(default)
    if isinstance(asset, str) and not asset.startswith(("http://", "https://")):
        return collection.add_asset(base_path / asset)
    return collection.add_asset(asset)


def load_resources_files_to_collection(base_path: Path) -> Collection:
    collection = Collection()
    for path in base_path.rglob("*.scp"):
        collection.load_from_scp(path)
    collection.load_from_source(base_path)
    return collection



from sonolus.backend.place import BlockPlace
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.value import Value
from sonolus.script.num import Num


@meta_fn
def deref[T: Value](block: Num, offset: Num, type_: type[T]) -> T:
    block = Num._accept_(block)
    offset = Num._accept_(offset)
    type_ = validate_value(type_)._as_py_()
    if block._is_py_():
        block = block._as_py_()
        if not isinstance(block, int):
            raise TypeError("block must be an integer")
    else:
        if not isinstance(block, Num):
            raise TypeError("block must be a Num")
        block = block.index()
    if offset._is_py_():
        offset = offset._as_py_()
        if not isinstance(offset, int):
            raise TypeError("offset must be an integer")
    else:
        if not isinstance(offset, Num):
            raise TypeError("offset must be a Num")
        offset = offset.index()
    if not (isinstance(type_, type) and issubclass(type_, Value)):
        raise TypeError("type_ must be a Value")
    return type_._from_place_(BlockPlace(block, offset))



from __future__ import annotations

from sonolus.build.collection import Asset
from sonolus.script.archetype import PlayArchetype, StandardImport


class Level:
    version = 1

    def __init__(
        self,
        *,
        name: str,
        title: str | None = None,
        rating: int = 0,
        artists: str = "Unknown",
        author: str = "Unknown",
        cover: Asset | None = None,
        bgm: Asset | None = None,
        data: LevelData,
    ) -> None:
        self.name = name
        self.title = title or name
        self.rating = rating
        self.artists = artists
        self.author = author
        self.cover = cover
        self.bgm = bgm
        self.data = data


class LevelData:
    bgm_offset: float
    entities: list[PlayArchetype]

    def __init__(self, bgm_offset: float, entities: list[PlayArchetype]) -> None:
        self.bgm_offset = bgm_offset
        self.entities = entities


class BpmChange(PlayArchetype):
    name = "#BPM_CHANGE"

    beat: StandardImport.Beat
    bpm: StandardImport.Bpm


class TimescaleChange(PlayArchetype):
    name = "#TIMESCALE_CHANGE"

    beat: StandardImport.Beat
    timescale: StandardImport.Timescale



from sonolus.backend.flow import BasicBlock, traverse_cfg_preorder
from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet
from sonolus.backend.passes import CompilerPass
from sonolus.backend.place import BlockPlace, TempBlock


class AllocateBasic(CompilerPass):
    def run(self, entry: BasicBlock):
        offsets = {}
        index = 16

        def process(stmt):
            nonlocal index
            match stmt:
                case int():
                    return stmt
                case IRConst():
                    return stmt
                case IRPureInstr(op=op, args=args):
                    return IRPureInstr(
                        op=op,
                        args=[process(arg) for arg in args],
                    )
                case IRInstr(op=op, args=args):
                    return IRInstr(
                        op=op,
                        args=[process(arg) for arg in args],
                    )
                case IRGet(place=place):
                    return IRGet(place=process(place))
                case IRSet(place=place, value=value):
                    return IRSet(place=process(place), value=process(value))
                case BlockPlace() as place:
                    if isinstance(place.block, TempBlock):
                        if place.block not in offsets:
                            offsets[place.block] = index
                            index += place.block.size
                        return BlockPlace(10000, process(place.index), place.offset + offsets[place.block])
                    return BlockPlace(
                        process(place.block) if isinstance(place.block, BlockPlace) else place.block,
                        process(place.index),
                        process(place.offset),
                    )
                case _:
                    raise NotImplementedError

        for block in traverse_cfg_preorder(entry):
            block.statements = [process(statement) for statement in block.statements]
            block.test = process(block.test)

        return entry
